__all__ = ("Pk2D", "parse_pk2d", "parse_pk",)

import warnings

import numpy as np

from . import (
    CCLObject, DEFAULT_POWER_SPECTRUM, UnlockInstance, check, get_pk_spline_a,
    get_pk_spline_lk, lib, unlock_instance)
from . import CCLWarning, CCLError
from .pyutils import _get_spline1d_arrays, _get_spline2d_arrays


class Pk2D(CCLObject):
    """A power spectrum class holding the information needed to reconstruct an
    arbitrary function of wavenumber and scale factor.

    .. note::

        The ``Pk2D`` class is a wrapper around CCL's bicubic interpolators.
        Addition, subtraction, multiplication, division returning new objects
        or in-place is supported between ``Pk2D`` objects and other ``Pk2D``
        objects, integers, or floats. When the second object is also of type
        ``Pk2D``, the a- and k-range changes to the most restrictive range.
        Exponentiation is also supported for integers and floats.

    .. note::

        The power spectrum can be evaluated by directly calling the instance
        ``pk(k, a)``. This is vectorized in both ``k`` and ``a``.

    Args:
        a_arr (`array`): 
            An array holding values of the scale factor
        lk_arr (`array`):
            An array holding values of the natural logarithm of the wavenumber
            (in :math:`\\mathrm{Mpc}^{-1}`).
        pk_arr (`array`) :
            A 2-D array of shape ``(na, nk)``, of the values of the power spectrum
            at ``a_arr`` and ``lk_arr``. Input array could be flattened, provided
            that its size is ``nk*na``. The array can hold the values of the
            natural logarithm of the power spectrum, depending on the value of
            ``is_logp``. Users must ensure that the power spectrum is sufficiently
            well sampled (i.e. the resolution of `a_arr` and `lk_arr` is high enough
            to sample the main features in the power spectrum). CCL uses bicubic
            interpolation to evaluate the power spectrum at any intermediate point
            in k and a.
        extrap_order_lok (:obj:`int`):  ``{0, 1, 2}``.
            Extrapolation order to be used on k-values below the minimum
            the splines. Note that extrapolation is either in :math:`\\log(P(k))`
            or in :math:`P(k)`, depending on the value of ``is_logp``.
        extrap_order_hik (:obj:`int`) :  ``{0, 1, 2}``.
            Extrapolation order to be used on k-values above the maximum of the
            splines. Note that extrapolation is either in :math:`\\log(P(k))` or
            in :math:`P(k)`, depending on the value of ``is_logp``.
        is_logp (:obj:`bool`):
            If True, ``pkfunc``/``pkarr`` return/hold the natural logarithm of the
            power spectrum. Otherwise, the true value of the power spectrum is
            expected. If ``is_logp`` is ``True``, arrays are interpolate in
            log-space.

    .. automethod:: __call__
    """ # noqa E501
    from ._core.repr_ import build_string_Pk2D as __repr__

    def __init__(self, *, a_arr=None, lk_arr=None, pk_arr=None,
                 is_logp=True, extrap_order_lok=1, extrap_order_hik=2):
        # Make sure input makes sense
        if (a_arr is None) or (lk_arr is None) or (pk_arr is None):
            raise ValueError("If you do not provide a function, "
                             "you must provide arrays")

        # Check that `a` is a monotonically increasing array.
        if not (np.diff(a_arr) > 0).all():
            raise ValueError("Input scale factor array in `a_arr` is not "
                             "monotonically increasing.")

        pkflat = pk_arr.flatten()
        # Check dimensions make sense
        if len(pkflat) != len(a_arr)*len(lk_arr):
            raise ValueError("Size of input arrays is inconsistent")

        status = 0
        self.psp, status = lib.set_pk2d_new_from_arrays(lk_arr, a_arr, pkflat,
                                                        int(extrap_order_lok),
                                                        int(extrap_order_hik),
                                                        int(is_logp), status)
        check(status)

    @classmethod
    def from_function(cls, pkfunc, *, is_logp=True,
                      spline_params=None,
                      extrap_order_lok=1, extrap_order_hik=2):
        """ Generates a `Pk2D` object from a function that calculates a power
        spectrum.

        Args:
            pkfunc (:obj:`callable`):
                Function with signature ``f(k, a)`` which takes vectorized
                input in ``k`` (wavenumber in :math:`\\mathrm{Mpc}^{-1}`)
                and a scale factor`a``, and returns the value of the
                corresponding quantity. ``pkfunc`` will be sampled at the
                values of ``k`` and ``a`` used internally by CCL to store the
                linear and non-linear power spectra.
            spline_params (:obj:`~pyccl._core.parameters.parameters_base.SplineParameters`):
                optional spline parameters. Used to determine sampling rates
                in scale factor and wavenumber. Custom parameters can be passed
                via the :class:`~pyccl.cosmology.Cosmology` object with
                ``cosmo.cosmo.spline_params`` (C API), or with an instance of
                ``ccl.parameters.SplineParameters`` (Python API). If ``None``, it
                defaults to the global accuracy parameters in CCL at the moment
                this function is called.

        Returns:
            :class:`~pyccl.pk2d.Pk2D`. Power spectrum object.
        """ # noqa E501
        if spline_params is None:
            from . import spline_params
        # Set k and a sampling from CCL parameters
        a_arr = get_pk_spline_a(spline_params=spline_params)
        lk_arr = get_pk_spline_lk(spline_params=spline_params)

        # Compute power spectrum on 2D grid
        pk_arr = np.array([pkfunc(k=np.exp(lk_arr), a=a) for a in a_arr])

        return cls(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                   is_logp=is_logp, extrap_order_lok=extrap_order_lok,
                   extrap_order_hik=extrap_order_hik)

    def __eq__(self, other):
        # Check object id.
        if self is other:
            return True
        # Check the object class.
        if type(self) is not type(other):
            return False
        # If the objects contain no data, return early.
        if not (self or other):
            return True
        # Check extrapolation orders.
        if not (self.extrap_order_lok == other.extrap_order_lok
                and self.extrap_order_hik == other.extrap_order_hik):
            return False
        # Check the individual splines.
        a1, lk1, pk1 = self.get_spline_arrays()
        a2, lk2, pk2 = other.get_spline_arrays()
        return ((a1 == a2).all() and (lk1 == lk2).all()
                and np.array_equal(pk1, pk2))

    def __hash__(self):
        return hash(repr(self))

    @property
    def has_psp(self):
        return 'psp' in vars(self)

    @property
    def extrap_order_lok(self):
        return self.psp.extrap_order_lok if self else None

    @property
    def extrap_order_hik(self):
        return self.psp.extrap_order_hik if self else None

    @classmethod
    def from_model(cls, cosmo, model):
        """:class:`Pk2D` constructor returning the power spectrum
        associated with a given numerical model.

        Arguments:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            model (:obj:`str`): Model to use. These models allowed:

                  - ``'bbks'`` (`Bardeen et al. 1986 <https://ui.adsabs.harvard.edu/abs/1986ApJ...304...15B/abstract>`_).
                  - ``'eisenstein_hu'`` (`Eisenstein & Hu 1997 <https://arxiv.org/abs/astro-ph/9709112>`_).
                  - ``'eisenstein_hu_nowiggles'`` (`Eisenstein & Hu 1997 <https://arxiv.org/abs/astro-ph/9709112>`_, no-wiggles version).

        Returns:
            :class:`~pyccl.pk2d.Pk2D`. The power spectrum of the input model.
        """  # noqa E501

        pk2d = Pk2D.__new__(cls)
        status = 0
        if model == 'bbks':
            cosmo.compute_growth()
            ret = lib.compute_linpower_bbks(cosmo.cosmo, status)
        elif model == 'eisenstein_hu':
            cosmo.compute_growth()
            ret = lib.compute_linpower_eh(cosmo.cosmo, 1, status)
        elif model == 'eisenstein_hu_nowiggles':
            cosmo.compute_growth()
            ret = lib.compute_linpower_eh(cosmo.cosmo, 0, status)
        else:
            raise ValueError(f"Invalid model {model}.")

        if np.ndim(ret) == 0:
            status = ret
        else:
            with UnlockInstance(pk2d):
                pk2d.psp, status = ret

        check(status, cosmo)
        return pk2d

    def apply_halofit(self, cosmo):
        """Apply the "HALOFIT" transformation of
        `Takahashi et al. 2012 <https://arxiv.org/abs/1208.2701>`_ on the linear
        power spectrum represented by this :class:`Pk2D` object, and return the
        result as a :class:`Pk2D` object.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.

        Return:
            :class:`Pk2D` object containing the non-linear power spectrum after
            applying HALOFIT.
        """ # noqa 501

        if cosmo["wa"] != 0:
            # HALOFIT translates (w0, wa) to a w0_eff. This requires computing
            # the comoving distance to the CMB, which requires the background
            # splines being sampled to sufficiently high redshifts.
            cosmo.compute_distances()
            _, a = _get_spline1d_arrays(cosmo.cosmo.data.achi)
            if min(a) > 1/(1 + 3000):
                raise CCLError("Comoving distance spline does not cover "
                               "sufficiently high redshifts for HALOFIT. "
                               "HALOFIT translates (w0, wa) to a w0_eff. This "
                               "requires computing the comoving distance to "
                               "the CMB, which requires the background "
                               "splines being sampled to sufficiently high "
                               "redshifts. If using the calculator mode, "
                               "check the support of the background data.")

        pk2d = Pk2D.__new__(Pk2D)
        status = 0
        ret = lib.apply_halofit(cosmo.cosmo, self.psp, status)
        if np.ndim(ret) == 0:
            status = ret
        else:
            with UnlockInstance(pk2d):
                pk2d.psp, status = ret
        check(status, cosmo)
        return pk2d

    def __call__(self, k, a, cosmo=None, *, derivative=False):
        """Evaluate the power spectrum or its logarithmic derivative at
        a single value of the scale factor.

        Arguments
        ---------
        k : :obj:`float` or `array`
            Wavenumber value(s) in units of :math:`{\\ rm Mpc}^{-1}`.
        a : :obj:`float` or `array`
            Value of the scale factor
        cosmo : :class:`~pyccl.cosmology.Cosmology`
            Cosmology object. Used to evaluate the power spectrum outside
            of the interpolation range in ``a``, thorugh the linear growth
            factor. If ``cosmo`` is ``None``, attempting to evaluate the power
            spectrum outside of the interpolation range will raise an error.
        derivative : :obj:`bool`
            If ``False``, evaluate the power spectrum. If ``True``, evaluate
            the logarithmic derivative of the power spectrum,
            :math:`d\\log P(k)/d\\log k`.

        Returns
        -------
        P(k, a) : (:obj:`float` or `array`)
            Value(s) of the power spectrum. or its derivative.
        """
        # determine if logarithmic derivative is needed
        if not derivative:
            eval_func = lib.pk2d_eval_multi
        else:
            eval_func = lib.pk2d_der_eval_multi

        # handle scale factor extrapolation
        if cosmo is None:
            cosmo = self.__call__._cosmo
            self.psp.extrap_linear_growth = 404  # flag no extrapolation
        else:
            cosmo.compute_growth()  # growth factors for extrapolation
            self.psp.extrap_linear_growth = 401  # flag extrapolation

        a_use = np.atleast_1d(a).astype(float)
        k_use = np.atleast_1d(k).astype(float)
        lk_use = np.log(k_use)

        status = 0
        out = np.zeros([len(a_use), len(k_use)])
        for ia, aa in enumerate(a_use):
            f, status = eval_func(self.psp, lk_use, aa,
                                  cosmo.cosmo, k_use.size, status)

            # Catch scale factor extrapolation bounds error.
            if status == lib.CCL_ERROR_SPLINE_EV:
                raise ValueError(
                    "Pk2D evaluation scale factor is outside of the "
                    "interpolation range. To extrapolate, pass a Cosmology.")
            check(status, cosmo)
            out[ia] = f

        if np.ndim(k) == 0:
            out = np.squeeze(out, axis=-1)
        if np.ndim(a) == 0:
            out = np.squeeze(out, axis=0)
        return out

    # Save a dummy cosmology as an attribute of the `__call__` method
    # so we don't have to initialize one every time no `cosmo` is passed.
    # This is gentle with memory too, as `free` does not work for an empty
    # cosmology.
    __call__._cosmo = type("Dummy", (object,), {"cosmo": lib.cosmology()})()

    def copy(self):
        """Return a copy of this Pk2D object."""
        if not self:
            return Pk2D.__new__(Pk2D)
        return self + 0

    def get_spline_arrays(self):
        """Get the spline data arrays internally stored by this object to
        interpolate the power spectrum.

        Returns:
            Tuple containing

            - a_arr: Array of scale factors.
            - lk_arr: Array of natural logarithm of wavenumber k.
            - pk_arr: Array of the power spectrum :math:`P(k, a)`. The shape
              is ``(a_arr.size, lk_arr.size)``.
        """
        if not self:
            raise ValueError("Pk2D object does not have data.")

        a_arr, lk_arr, pk_arr = _get_spline2d_arrays(self.psp.fka)
        if self.psp.is_log:
            pk_arr = np.exp(pk_arr)

        return a_arr, lk_arr, pk_arr

    def __del__(self):
        """Free memory associated with this Pk2D structure."""
        if self:
            lib.f2d_t_free(self.psp)

    def __bool__(self):
        return self.has_psp

    def __contains__(self, other):
        if not (self.psp.lkmin <= other.psp.lkmin
                and self.psp.lkmax >= other.psp.lkmax
                and self.psp.amin <= other.psp.amin
                and self.psp.amax >= other.psp.amax):
            return False
        return True

    def _get_binary_operator_arrays(self, other):
        if not (self and other):
            raise ValueError("Pk2D object does not have data.")
        if self not in other:
            raise ValueError(
                "The 2nd operand has its data defined over a smaller range "
                "than the 1st operand. To avoid extrapolation, this operation "
                "is forbidden. If you want to operate on the smaller support, "
                "try swapping the operands.")

        a_arr_a, lk_arr_a, pk_arr_a = self.get_spline_arrays()
        a_arr_b, lk_arr_b, pk_arr_b = other.get_spline_arrays()
        if not (a_arr_a.size == a_arr_b.size
                and lk_arr_a.size == lk_arr_b.size
                and np.allclose(a_arr_a, a_arr_b)
                and np.allclose(lk_arr_a, lk_arr_b)):
            warnings.warn(
                "Operands defined over different ranges. "
                "The result will be interpolated and clipped to "
                f"{self.psp.lkmin} <= log k <= {self.psp.lkmax} and "
                f"{self.psp.amin} <= a <= {self.psp.amax}.", CCLWarning)
            pk_arr_b = other(np.exp(lk_arr_a), a_arr_a)

        return a_arr_a, lk_arr_a, pk_arr_a, pk_arr_b

    def __add__(self, other):
        """Adds two Pk2D instances.

        The a and k ranges of the 2nd operand need to be the same or smaller
        than the 1st operand.
        The returned Pk2D object uses the same a and k arrays as the first
        operand.
        """
        if isinstance(other, (float, int)):
            a_arr_a, lk_arr_a, pk_arr_a = self.get_spline_arrays()
            pk_arr_new = pk_arr_a + other
        elif isinstance(other, Pk2D):
            a_arr_a, lk_arr_a, pk_arr_a, pk_arr_b = \
                self._get_binary_operator_arrays(other)
            pk_arr_new = pk_arr_a + pk_arr_b
        else:
            raise TypeError("Addition of Pk2D is only defined for "
                            "floats, ints, and Pk2D objects.")

        logp = np.all(pk_arr_new > 0)
        if logp:
            pk_arr_new = np.log(pk_arr_new)

        new = Pk2D(a_arr=a_arr_a, lk_arr=lk_arr_a, pk_arr=pk_arr_new,
                   is_logp=logp,
                   extrap_order_lok=self.extrap_order_lok,
                   extrap_order_hik=self.extrap_order_hik)

        return new

    def __mul__(self, other):
        """Multiply two Pk2D instances.

        The a and k ranges of the 2nd operand need to be the same or smaller
        than the 1st operand.
        The returned Pk2D object uses the same a and k arrays as the first
        operand.
        """
        if isinstance(other, (float, int)):
            a_arr_a, lk_arr_a, pk_arr_a = self.get_spline_arrays()
            pk_arr_new = other * pk_arr_a
        elif isinstance(other, Pk2D):
            a_arr_a, lk_arr_a, pk_arr_a, pk_arr_b = \
                self._get_binary_operator_arrays(other)
            pk_arr_new = pk_arr_a * pk_arr_b
        else:
            raise TypeError("Multiplication of Pk2D is only defined for "
                            "floats, ints, and Pk2D objects.")

        logp = np.all(pk_arr_new > 0)
        if logp:
            pk_arr_new = np.log(pk_arr_new)

        new = Pk2D(a_arr=a_arr_a, lk_arr=lk_arr_a, pk_arr=pk_arr_new,
                   is_logp=logp,
                   extrap_order_lok=self.extrap_order_lok,
                   extrap_order_hik=self.extrap_order_hik)
        return new

    def __pow__(self, exponent):
        """Take a Pk2D instance to a power.
        """
        if not isinstance(exponent, (float, int)):
            raise TypeError(
                "Exponentiation of Pk2D is only defined for floats and ints.")
        a_arr_a, lk_arr_a, pk_arr_a = self.get_spline_arrays()
        if np.any(pk_arr_a < 0) and exponent % 1 != 0:
            warnings.warn(
                "Taking a non-positive Pk2D object to a non-integer "
                "power may lead to unexpected results", CCLWarning)

        pk_arr_new = pk_arr_a**exponent

        logp = np.all(pk_arr_new > 0)
        if logp:
            pk_arr_new = np.log(pk_arr_new)

        new = Pk2D(a_arr=a_arr_a, lk_arr=lk_arr_a, pk_arr=pk_arr_new,
                   is_logp=logp,
                   extrap_order_lok=self.extrap_order_lok,
                   extrap_order_hik=self.extrap_order_hik)

        return new

    def __sub__(self, other):
        return self + (-1)*other

    def __truediv__(self, other):
        return self * other**(-1)

    __radd__ = __add__

    __rmul__ = __mul__

    def __rsub__(self, other):
        return other + (-1)*self

    def __rtruediv__(self, other):
        return other * self**(-1)

    @unlock_instance
    def __iadd__(self, other):
        self = self + other
        return self

    @unlock_instance
    def __imul__(self, other):
        self = self * other
        return self

    @unlock_instance
    def __isub__(self, other):
        self = self - other
        return self

    @unlock_instance
    def __itruediv__(self, other):
        self = self / other
        return self

    @unlock_instance
    def __ipow__(self, other):
        self = self**other
        return self


def parse_pk2d(cosmo, p_of_k_a=DEFAULT_POWER_SPECTRUM, *, is_linear=False):
    """ Return the C-level `f2d` spline associated with a
    :class:`Pk2D` object.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        p_of_k_a (:class:`Pk2D` or :obj:`str`): if a
            :class:`Pk2D` object, its `f2d` spline will be used. If
            a string, the linear or non-linear power spectrum stored
            by ``cosmo`` under this name will be used. Defaults to the
            matter power spectrum stored in `cosmo`.
        is_linear (:obj:`bool`): if ``True``, and if ``p_of_k_a`` is a
            string or ``None``, the linear version of the corresponding
            power spectrum will be used (otherwise it'll be the
            non-linear version).
    """
    if isinstance(p_of_k_a, Pk2D):
        psp = p_of_k_a.psp
    else:
        if isinstance(p_of_k_a, str):
            name = p_of_k_a
        else:
            raise ValueError("p_of_k_a must be a pyccl.Pk2D object or "
                             "a string")

        if is_linear:
            cosmo.compute_linear_power()
            pk = cosmo.get_linear_power(name)
        else:
            cosmo.compute_nonlin_power()
            pk = cosmo.get_nonlin_power(name)
        psp = pk.psp
    return psp


def parse_pk(cosmo, p_of_k_a=None):
    """Helper to retrieve the right :class:`Pk2D` object.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`):
            3D Power spectrum to return. If a `Pk2D`, it is just returned.
            If `None` or `"linear"`, the linear power spectrum stored in
            ``cosmo`` is returned. If `"nonlinear"`, the nonlinear matter
            power spectrum stored in ``cosmo`` is returned.

    Returns:
        :class:`Pk2D` object corresponding to ``p_of_k_a``.
    """
    if not (p_of_k_a is None or isinstance(p_of_k_a, (str, Pk2D))):
        raise TypeError("p_of_k_a must be None, 'linear', 'nonlinear', Pk2D.")

    if isinstance(p_of_k_a, Pk2D):
        return p_of_k_a
    elif p_of_k_a is None or p_of_k_a == "linear":
        cosmo.compute_linear_power()
        return cosmo.get_linear_power()
    elif p_of_k_a == "nonlinear":
        cosmo.compute_nonlin_power()
        return cosmo.get_nonlin_power()



import numpy as np
import pyccl as ccl

BCM_TOLERANCE = 1e-4


def test_bcm():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.2e-9,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.0,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    data = np.loadtxt("./benchmarks/data/w_baryonspk_nl.dat")
    data_nobar = np.loadtxt("./benchmarks/data/wo_baryonspk_nl.dat")

    k = data[:, 0] * cosmo['h']
    a = 1

    bar = ccl.BaryonsSchneider15(log10Mc=14.)
    fbcm = bar.boost_factor(cosmo, k, a)
    err = np.abs(data[:, 1]/data_nobar[:, 1]/fbcm - 1)
    assert np.allclose(err, 0, atol=BCM_TOLERANCE, rtol=0)

    cosmo.compute_nonlin_power()
    pk_nobar = cosmo.get_nonlin_power()
    pk_wbar = bar.include_baryonic_effects(cosmo, pk_nobar)
    ratio = pk_wbar(k, a)/pk_nobar(k, a)
    err = np.abs(data[:, 1]/data_nobar[:, 1]/ratio - 1)
    assert np.allclose(err, 0, atol=BCM_TOLERANCE, rtol=0)



__all__ = ("HMIngredients", "Concentration", "MassFunc", "HaloBias",
           "get_delta_c", )

from abc import abstractmethod

import numpy as np

from .. import CCLAutoRepr, CCLNamedClass, lib, check
from .. import physical_constants as const


def get_delta_c(cosmo, a, kind='EdS'):
    """Returns the linear collapse threshold.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        a (:obj:`float` or `array`): scale factor.
        kind (:obj:`str`): prescription to use. Should be one of

            * 'EdS': the SC prediction in Einstein de-Sitter, :math:`\\delta_c=(3/20)(12\\pi)^{2/3}`.
            * 'EdS_approx': a common approximation to the EdS result :math:`\\delta_c=1.686`.
            * 'NakamuraSuto97': the prescription from `Nakamura & Suto 1997 <https://arxiv.org/abs/astro-ph/9612074>`_.
            * 'Mead16': the prescription from `Mead et al. 2016 <https://arxiv.org/abs/1602.02154>`_.

    Returns:
        (:obj:`float` or `array`): linear collapse threshold.
    """  # noqa
    # This is the linear collapse threshold in Einstein de-Sitter:
    # delta_c = 3/20*(12*pi)^(2/3)
    dc0 = 1.68647019984

    if kind == 'EdS':
        return dc0
    elif kind == 'EdS_approx':
        return 1.686
    elif kind == 'NakamuraSuto97':
        Om = cosmo.omega_x(a, 'matter')
        return dc0*(1+0.012299*np.log10(Om))
    elif kind == 'Mead16':
        Om = cosmo.omega_x(a, 'matter')
        s8 = cosmo.sigma8()*cosmo.growth_factor(a)
        facs8 = (1.59+0.0314*np.log(s8))
        facOm = (1+0.0123*np.log10(Om))
        return facs8*facOm
    else:
        raise ValueError(f"Unknown threshold kind {kind}")


class HMIngredients(CCLAutoRepr, CCLNamedClass):
    """Base class for halo model ingredients."""
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict",)

    def __init__(self, *, mass_def, mass_def_strict=True):
        # Check mass definition consistency.
        from .massdef import MassDef
        mass_def = MassDef.create_instance(mass_def)
        self.mass_def_strict = mass_def_strict
        self._check_mass_def(mass_def)
        self.mass_def = mass_def
        self._setup()

    @property
    @abstractmethod
    def _mass_def_strict_always(self) -> bool:
        """Property that dictates whether ``mass_def_strict`` can be set
        as False on initialization.

        Some models are set up in a way so that the set of fitted parameters
        depends on the mass definition, (i.e. no one universal model exists
        to cover all cases). Setting this to True fixes ``mass_def_strict``
        to True irrespective of what the user passes.

        Set this propery to False to allow users to override strict checks.
        """

    @abstractmethod
    def _check_mass_def_strict(self, mass_def) -> bool:
        """Check if this class is defined for mass definition ``mass_def``."""

    def _setup(self) -> None:
        """ Use this function to initialize any internal attributes
        of this object. This function is called at the very end of the
        constructor call.
        """

    def _check_mass_def(self, mass_def) -> None:
        """ Return False if the input mass definition agrees with
        the definitions for which this mass function parametrization
        works. True otherwise. This function gets called at the
        start of the constructor call.

        Args:
            mass_def (:class:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.
        """
        classname = self.__class__.__name__
        msg = f"{classname} is not defined for {mass_def.name} masses"

        if self._check_mass_def_strict(mass_def):
            # Passed mass is incompatible with model.

            if self._mass_def_strict_always:
                # Class has no universal model and mass is incompatible.
                raise ValueError(
                    f"{msg} and this requirement cannot be relaxed.")

            if self.mass_def_strict:
                # Strict mass_def check enabled and mass is incompatible.
                raise ValueError(
                    f"{msg}. To relax this check set `mass_def_strict=False`.")

    def _get_logM_sigM(self, cosmo, M, a, *, return_dlns=False):
        """Compute ``logM``, ``sigM``, and (optionally) ``dlns_dlogM``."""
        cosmo.compute_sigma()  # initialize sigma(M) splines if needed
        logM = np.log10(M)

        # sigma(M)
        status = 0
        sigM, status = lib.sigM_vec(cosmo.cosmo, a, logM, len(logM), status)
        check(status, cosmo=cosmo)
        if not return_dlns:
            return logM, sigM

        # dlogsigma(M)/dlog10(M)
        dlns_dlogM, status = lib.dlnsigM_dlogM_vec(cosmo.cosmo, a, logM,
                                                   len(logM), status)
        check(status, cosmo=cosmo)
        return logM, sigM, dlns_dlogM


class MassFunc(HMIngredients):
    """This class enables the calculation of halo mass functions.
    We currently assume that all mass functions can be written as

    .. math::
        \\frac{dn}{d\\log_{10}M} = f(\\sigma_M)\\,\\frac{\\rho_M}{M}\\,
        \\frac{d\\log \\sigma_M}{d\\log_{10} M}

    where :math:`\\sigma_M^2` is the overdensity variance on spheres with a
    radius given by the Lagrangian radius for mass M.

    * Subclasses implementing analytical mass function parametrizations
      can be created by overriding the ``_get_fsigma`` method.

    * Subclasses may have particular implementations of
      ``_check_mass_def_strict`` to ensure consistency of the halo mass
      definition.

    * Subclasses for parametrizations that cannot be written in terms of
      :math:`\\sigma_M` can simply overload the :meth:`__call__` method.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object or a name string.
        mass_def_strict (:obj:`bool`): if False, consistency of the mass
            definition will be ignored.

    .. automethod:: __call__
    """
    _mass_def_strict_always = False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        """ Get the :math:`f(\\sigma_M)` function for this mass function
        object (see description of this class for details).

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            sigM (:obj:`float` or `array`): standard deviation in the
                overdensity field on the scale of this halo.
            a (:obj:`float`): scale factor.
            lnM (:obj:`float` or `array`): natural logarithm of the
                halo mass in units of M_sun (provided in addition
                to sigM for convenience in some mass function
                parametrizations).

        Returns:
            (:obj:`float` or `array`): :math:`f(\\sigma_M)` function.
        """

    def __call__(self, cosmo, M, a):
        """ Returns the mass function for input parameters.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            M (:obj:`float` or `array`): halo mass.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): mass function \
                :math:`dn/d\\log_{10}M` in units of Mpc^-3 (comoving).
        """
        M_use = np.atleast_1d(M)
        logM, sigM, dlns_dlogM = self._get_logM_sigM(
            cosmo, M_use, a, return_dlns=True)

        rho = (const.RHO_CRITICAL * cosmo['Omega_m'] * cosmo['h']**2)
        f = self._get_fsigma(cosmo, sigM, a, 2.302585092994046 * logM)
        mf = f * rho * dlns_dlogM / M_use
        if np.ndim(M) == 0:
            return mf[0]
        return mf


class HaloBias(HMIngredients):
    """This class enables the calculation of halo bias functions.
    We currently assume that all halo bias functions can be written
    as functions that depend on :math:`M` only through :math:`\\sigma_M`
    (where :math:`\\sigma_M^2` is the overdensity variance on
    spheres with a radius given by the Lagrangian radius for
    mass :math:`M`). All sub-classes implementing specific
    parametrizations can therefore be simply created by replacing
    this class' ``_get_bsigma`` method. New classes departing
    from this paradigm can simply overload the
    :meth:`__call__` method.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object or a name string.
        mass_def_strict (:obj:`bool`): if False, consistency of the mass
            definition will be ignored.

    .. automethod:: __call__
    """
    _mass_def_strict_always = False

    def _get_bsigma(self, cosmo, sigM, a):
        """ Get the halo bias as a function of sigmaM.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            sigM (:obj:`float` or `array`): standard deviation in the
                overdensity field on the scale of this halo.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): f(sigma_M) function.
        """

    def __call__(self, cosmo, M, a):
        """Returns the halo bias for input parameters.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            M (:obj:`float` or `array`): halo mass.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): halo bias.
        """
        M_use = np.atleast_1d(M)
        logM, sigM = self._get_logM_sigM(cosmo, M_use, a)
        b = self._get_bsigma(cosmo, sigM, a)
        if np.ndim(M) == 0:
            return b[0]
        return b


class Concentration(HMIngredients):
    """
    This class enables the calculation of halo concentrations.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object or a name string.

    .. automethod:: __call__
    """
    _mass_def_strict_always = True

    def __init__(self, *, mass_def):
        super().__init__(mass_def=mass_def, mass_def_strict=True)

    def _concentration(self, cosmo, M, a):
        """Implementation of the c(M) relation."""

    def __call__(self, cosmo, M, a):
        """Returns the concentration for input parameters.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            M (:obj:`float` or `array`): halo mass.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): concentration.
        """
        M_use = np.atleast_1d(M)
        c = self._concentration(cosmo, M_use, a)
        if np.ndim(M) == 0:
            return c[0]
        return c



__all__ = ("CorrelationMethods", "CorrelationTypes", "correlation",
           "correlation_3d", "correlation_multipole", "correlation_3dRsd",
           "correlation_3dRsd_avgmu", "correlation_pi_sigma",)

from enum import Enum
import numpy as np
from . import DEFAULT_POWER_SPECTRUM, check, lib


class CorrelationMethods(Enum):
    """Choices of algorithms used to compute correlation functions:

    - 'Bessel' is a direct integration using Bessel functions.
    - 'FFTLog' is fast using a fast Fourier transform.
    - 'Legendre' uses a sum over Legendre polynomials.
    """
    FFTLOG = "fftlog"
    BESSEL = "bessel"
    LEGENDRE = "legendre"


class CorrelationTypes(Enum):
    """Correlation function types.
    """
    NN = "NN"
    NG = "NG"
    GG_PLUS = "GG+"
    GG_MINUS = "GG-"


correlation_methods = {
    'fftlog': lib.CCL_CORR_FFTLOG,
    'bessel': lib.CCL_CORR_BESSEL,
    'legendre': lib.CCL_CORR_LGNDRE,
}

correlation_types = {
    'NN': lib.CCL_CORR_GG,
    'NG': lib.CCL_CORR_GL,
    'GG+': lib.CCL_CORR_LP,
    'GG-': lib.CCL_CORR_LM,
}


def correlation(cosmo, *, ell, C_ell, theta, type='NN', method='fftlog'):
    r"""Compute the angular correlation function.

    .. math::

        \xi^{ab}_\pm(\theta) =
        \sum_\ell\frac{2\ell+1}{4\pi}\,(\pm1)^{s_b}\,
        C^{ab\pm}_\ell\,d^\ell_{s_a,\pm s_b}(\theta)

    where :math:`\theta` is the angle between the two fields :math:`a` and
    :math:`b` with spins :math:`s_a` and :math:`s_b` after alignement of their
    tangential coordinate. :math:`d^\ell_{mm'}` are the Wigner-d matrices and
    we have defined the power spectra

    .. math::
        C^{ab\pm}_\ell \equiv
        (C^{a_Eb_E}_\ell \pm C^{a_Bb_B}_\ell)+i
        (C^{a_Bb_E}_\ell \mp C^{a_Eb_B}_\ell),

    which reduces to the :math:`EE` power spectrum when all :math:`B`-modes
    are 0.

    The different spin combinations are:

        * :math:`s_a=s_b=0` e.g. galaxy-galaxy, galaxy-:math:`\kappa`
          and :math:`\kappa`-:math:`\kappa`
        * :math:`s_a=2`, :math:`s_b=0` e.g. galaxy-shear, and :math:`\kappa`-shear
        * :math:`s_a=s_b=2` e.g. shear-shear.

    .. note::
        For scales smaller than :math:`\sim 0.1^{\circ}`, the input power
        spectrum should be sampled to sufficienly high :math:`\ell` to ensure
        the Hankel transform is well-behaved. The following spline parameters,
        related to ``FFTLog``-sampling may also be modified for accuracy:

            * ``ccl.spline_params.ELL_MIN_CORR``
            * ``ccl.spline_params.ELL_MAX_CORR``
            * ``ccl.spline_params.N_ELL_CORR``.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        ell (array): Multipoles corresponding to the input angular power
                          spectrum.
        C_ell (array): Input angular power spectrum.
        theta (:obj:`float` or `array`): Angular separation(s) at which to
            calculate the angular correlation function (in degrees).
        type (:obj:`str`): Type of correlation function. Choices: ``'NN'`` (0x0),
            ``'NG'`` (0x2), ``'GG+'`` (2x2, :math:`\xi_+`),
            ``'GG-'`` (2x2, :math:`\xi_-`), where numbers refer to the spins
            of the two quantities being cross-correlated (see Section 2.4.2 of
            the CCL paper). The naming system roughly follows the nomenclature
            used in `TreeCorr
            <https://rmjarvis.github.io/TreeCorr/_build/html/correlation2.html>`_.
        method (:obj:`str`): Method to compute the correlation function.
            Choices: ``'Bessel'`` (direct integration over Bessel function),
            ``'FFTLog'`` (fast integration with FFTLog), ``'Legendre'``
            (brute-force sum over Legendre polynomials).

    Returns:
        (:obj:`float` or `array`): Value(s) of the correlation function at the
        input angular separations.
    """ # noqa
    cosmo_in = cosmo
    cosmo = cosmo.cosmo
    status = 0
    method = method.lower()

    if type not in correlation_types:
        raise ValueError(f"Invalud correlation type {type}.")

    if method not in correlation_methods.keys():
        raise ValueError(f"Invalid correlation method {method}.")

    # Convert scalar input into an array
    if scalar := isinstance(theta, (int, float)):
        theta = np.array([theta, ])

    if np.all(np.array(C_ell) == 0):
        # short-cut and also avoid integration errors
        wth = np.zeros_like(theta)
    else:
        # Call correlation function
        wth, status = lib.correlation_vec(cosmo, ell, C_ell, theta,
                                          correlation_types[type],
                                          correlation_methods[method],
                                          len(theta), status)
    check(status, cosmo_in)
    if scalar:
        return wth[0]
    return wth


def correlation_3d(cosmo, *, r, a, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    r"""Compute the 3D correlation function:

    .. math::
        \xi(r)\equiv\frac{1}{2\pi^2}\int dk\,k^2\,P(k)\,j_0(kr).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        r (:obj:`float` or `array`): distance(s) at which to calculate the 3D
            correlation function (in Mpc).
        a (:obj:`float`): scale factor.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power spectrum
            to integrate. If a string, it must correspond to one of the
            non-linear power spectra stored in `cosmo` (e.g.
            `'delta_matter:delta_matter'`).

    Returns:
        Value(s) of the correlation function at the input distance(s).
    """ # noqa
    cosmo.compute_nonlin_power()

    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a)

    status = 0

    # Convert scalar input into an array
    if scalar := isinstance(r, (int, float)):
        r = np.array([r, ])

    # Call 3D correlation function
    xi, status = lib.correlation_3d_vec(cosmo, psp, a, r,
                                        len(r), status)
    check(status, cosmo_in)
    if scalar:
        return xi[0]
    return xi


def correlation_multipole(cosmo, *, r, a, beta, ell,
                          p_of_k_a=DEFAULT_POWER_SPECTRUM):
    r"""Compute the correlation function multipoles:

    .. math::
        \xi_\ell(r)\equiv\frac{i^\ell}{2\pi^2}\int dk\,k^2\,P(k)\,j_\ell(kr).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        r (:obj:`float` or `array`): distance(s) at which to calculate the 3D
            correlation function (in Mpc).
        a (:obj:`float`): scale factor.
        beta (:obj:`float`): growth rate divided by galaxy bias.
        ell (:obj:`int`) : the desired multipole
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power spectrum
            to integrate. If a string, it must correspond to one of the
            non-linear power spectra stored in `cosmo` (e.g.
            `'delta_matter:delta_matter'`).

    Returns:
        Value(s) of the correlation function at the input distance(s).
    """ # noqa
    cosmo.compute_nonlin_power()

    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a)

    status = 0

    # Convert scalar input into an array
    if scalar := isinstance(r, (int, float)):
        r = np.array([r, ])

    # Call 3D correlation function
    xis, status = lib.correlation_multipole_vec(cosmo, psp, a, beta, ell, r,
                                                len(r), status)
    check(status, cosmo_in)
    if scalar:
        return xis[0]
    return xis


def correlation_3dRsd(cosmo, *, r, a, mu, beta,
                      p_of_k_a=DEFAULT_POWER_SPECTRUM, use_spline=True):
    r"""
    Compute the 3D correlation function with linear RSDs using
    multipoles:

    .. math::
        \xi(r,\mu) = \sum_{\ell\in\{0,2,4\}}\xi_\ell(r)\,P_\ell(\mu)

    where :math:`P_\ell(\mu)` are the Legendre polynomials, and
    :math:`\xi_\ell(r)` are the correlation function multipoles.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        r (:obj:`float` or `array`): distance(s) at which to calculate the
            3D correlation function (in Mpc).
        a (:obj:`float`): scale factor.
        mu (:obj:`float`): cosine of the angle at which to calculate the 3D
            correlation function.
        beta (:obj:`float`): growth rate divided by galaxy bias.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power spectrum
            to integrate. If a string, it must correspond to one of the
            non-linear power spectra stored in `cosmo` (e.g.
            `'delta_matter:delta_matter'`).
        use_spline (:obj:`bool`): switch that determines whether the RSD correlation
            function is calculated using global splines of multipoles.

    Returns:
        Value(s) of the correlation function at the input distance(s) & angle.
    """ # noqa
    cosmo.compute_nonlin_power()

    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a)

    status = 0

    # Convert scalar input into an array
    if scalar := isinstance(r, (int, float)):
        r = np.array([r, ])

    # Call 3D correlation function
    xis, status = lib.correlation_3dRsd_vec(cosmo, psp, a, mu, beta, r,
                                            len(r), int(use_spline), status)
    check(status, cosmo_in)
    if scalar:
        return xis[0]
    return xis


def correlation_3dRsd_avgmu(cosmo, *, r, a, beta,
                            p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """
    Compute the 3D correlation function averaged over angles with
    RSDs. Equivalent to calling :func:`correlation_multipole`
    with ``ell=0``.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        r (:obj:`float` or `array`): distance(s) at which to calculate the 3D
            correlation function (in Mpc).
        a (:obj:`float`): scale factor.
        beta (:obj:`float`): growth rate divided by galaxy bias.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power spectrum
            to integrate. If a string, it must correspond to one of the
            non-linear power spectra stored in `cosmo` (e.g.
            `'delta_matter:delta_matter'`).

    Returns:
        Value(s) of the correlation function at the input distance(s) & angle.
    """ # noqa
    cosmo.compute_nonlin_power()

    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a)

    status = 0

    # Convert scalar input into an array
    if scalar := isinstance(r, (int, float)):
        r = np.array([r, ])

    # Call 3D correlation function
    xis, status = lib.correlation_3dRsd_avgmu_vec(cosmo, psp, a, beta, r,
                                                  len(r), status)
    check(status, cosmo_in)
    if scalar:
        return xis[0]
    return xis


def correlation_pi_sigma(cosmo, *, pi, sigma, a, beta,
                         use_spline=True, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    r"""
    Compute the 3D correlation in :math:`(\pi,\sigma)` space. This is
    just

    .. math::
        \xi(\pi,\sigma) = \xi(r=\sqrt{\pi^2+\sigma^2},\mu=\pi/r).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        pi (:obj:`float`): distance times cosine of the angle (in Mpc).
        sigma (:obj:`float` or `array`): distance(s) times sine of the angle
            (in Mpc).
        a (:obj:`float`): scale factor.
        beta (:obj:`float`): growth rate divided by galaxy bias.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power spectrum
            to integrate. If a string, it must correspond to one of the
            non-linear power spectra stored in `cosmo` (e.g.
            `'delta_matter:delta_matter'`).
        use_spline (:obj:`bool`): switch that determines whether the RSD correlation
            function is calculated using global splines of multipoles.

    Returns:
        Value(s) of the correlation function at the input pi and sigma.
    """ # noqa
    cosmo.compute_nonlin_power()

    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a)

    status = 0

    # Convert scalar input into an array
    if scalar := isinstance(sigma, (int, float)):
        sigma = np.array([sigma, ])

    # Call 3D correlation function
    xis, status = lib.correlation_pi_sigma_vec(cosmo, psp, a, beta, pi, sigma,
                                               len(sigma), int(use_spline),
                                               status)
    check(status, cosmo_in)
    if scalar:
        return xis[0]
    return xis



__all__ = ("HaloProfilePressureGNFW",)

import numpy as np

from ... import UnlockInstance
from . import HaloProfilePressure


class HaloProfilePressureGNFW(HaloProfilePressure):
    """ Generalized NFW electron pressure profile by
    `Arnaud et al. 2010 <https://arxiv.org/abs/0910.1234>`_.

    The parametrization is:

    .. math::

       P_e(r) = C\\times P_0 h_{70}^E (c_{500} x)^{-\\gamma}
       [1+(c_{500}x)^\\alpha]^{(\\gamma-\\beta)/\\alpha},

    where

    .. math::

       C = 1.65\\,h_{70}^2\\left(\\frac{H(z)}{H_0}\\right)^{8/3}
       \\left[\\frac{h_{70}\\tilde{M}_{500c}}
       {3\\times10^{14}\\,M_\\odot}\\right]^{2/3+\\alpha_{\\mathrm{P}}},

    :math:`x = r/\\tilde{r}_{500c}`, :math:`h_{70}=h/0.7`, and the
    exponent :math:`E` is -1 for SZ-based profile normalizations
    and -1.5 for X-ray-based normalizations. The biased mass
    :math:`\\tilde{M}_{500c}` is related to the true overdensity
    mass :math:`M_{500c}` via the mass bias parameter :math:`(1-b)`
    as :math:`\\tilde{M}_{500c}=(1-b)M_{500c}`. :math:`\\tilde{r}_{500c}`
    is the overdensity halo radius associated with :math:`\\tilde{M}_{500c}`
    (note the intentional tilde!), and the profile is defined for
    a halo overdensity :math:`\\Delta=500` with respect to the
    critical density.

    The default arguments (other than ``mass_bias``), correspond to the
    profile parameters used in the `Planck 2013 (XX)
    <https://arxiv.org/abs/1303.5080>`_ paper. The profile is
    calculated in physical (non-comoving) units of :math:`\\mathrm{eV/cm^3}`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_bias (:obj:`float`):
            The mass bias parameter :math:`1-b`.
        P0 (:obj:`float`):
            Profile normalization.
        c500 (:obj:`float`):
            Concentration parameter.
        alpha, beta, gamma (:obj:`float`):
            Profile shape parameters.
        alpha_P (:obj:`float`):
            Additional mass dependence exponent
        P0_hexp (:obj:`float`):
            Power of :math:`h` with which the normalization parameter scales.
            Equal to :math:`-1` for SZ-based normalizations,
            and :math:`-3/2` for X-ray-based normalizations.
        qrange (:obj:`tuple`):
            Tuple of two numbers denoting the limits of integration used when
            computing the Fourier-space profile template, in units of
            :math:`r_{\\mathrm{vir}}`.
        nq (:obj:`int`):
            Number of sampling points of the Fourier-space profile template.
        x_out (:obj:`float`):
            Profile threshold, in units of :math:`r_{\\mathrm{500c}}`.
            Defaults to :math:`+\\infty`.
    """
    __repr_attrs__ = __eq_attrs__ = (
        "mass_bias", "P0", "c500", "alpha", "alpha_P", "beta", "gamma",
        "P0_hexp", "qrange", "nq", "x_out", "mass_def", "precision_fftlog",)

    def __init__(self, *, mass_def, mass_bias=0.8, P0=6.41,
                 c500=1.81, alpha=1.33, alpha_P=0.12,
                 beta=4.13, gamma=0.31, P0_hexp=-1.,
                 qrange=(1e-3, 1e3), nq=128, x_out=np.inf):
        self.qrange = qrange
        self.nq = nq
        self.mass_bias = mass_bias
        self.P0 = P0
        self.c500 = c500
        self.alpha = alpha
        self.alpha_P = alpha_P
        self.beta = beta
        self.gamma = gamma
        self.P0_hexp = P0_hexp
        self.x_out = x_out

        # Interpolator for dimensionless Fourier-space profile
        self._fourier_interp = None
        super().__init__(mass_def=mass_def)

    def update_parameters(self, *, mass_bias=None, P0=None,
                          c500=None, alpha=None, beta=None, gamma=None,
                          alpha_P=None, P0_hexp=None, x_out=None):
        """Update any of the parameters associated with this profile.
        Any parameter set to ``None`` won't be updated.

        .. note::

            A change in ``alpha``, ``beta``, ``gamma``, ``c500``, or ``x_out``
            recomputes the Fourier-space template, which may be slow.

        Args:
            mass_bias (:obj:`float`):
                The mass bias parameter :math:`1-b`.
            P0 (:obj:`float`):
                Profile normalization.
            c500 (:obj:`float`):
                Concentration parameter.
            alpha, beta, gamma (:obj:`float`):
                Profile shape parameters.
            alpha_P (:obj:`float`):
                Additional mass dependence exponent
            P0_hexp (:obj:`float`):
                Power of :math:`h` with which the normalization parameter
                scales. Equal to :math:`-1` for SZ-based normalizations,
                and :math:`-3/2` for X-ray-based normalizations.
            x_out (:obj:`float`):
                Profile threshold, in units of :math:`r_{\\mathrm{500c}}`.
                Defaults to :math:`+\\infty`.
        """
        if mass_bias is not None:
            self.mass_bias = mass_bias
        if alpha_P is not None:
            self.alpha_P = alpha_P
        if P0 is not None:
            self.P0 = P0
        if P0_hexp is not None:
            self.P0_hexp = P0_hexp

        # Check if we need to recompute the Fourier profile.
        re_fourier = False
        if alpha is not None and alpha != self.alpha:
            re_fourier = True
            self.alpha = alpha
        if beta is not None and beta != self.beta:
            re_fourier = True
            self.beta = beta
        if gamma is not None and gamma != self.gamma:
            re_fourier = True
            self.gamma = gamma
        if c500 is not None and c500 != self.c500:
            re_fourier = True
            self.c500 = c500
        if x_out is not None and x_out != self.x_out:
            re_fourier = True
            self.x_out = x_out

        if re_fourier and (self._fourier_interp is not None):
            self._fourier_interp = self._integ_interp()

    def _form_factor(self, x):
        # Scale-dependent factor of the GNFW profile.
        f1 = (self.c500*x)**(-self.gamma)
        exponent = -(self.beta-self.gamma)/self.alpha
        f2 = (1+(self.c500*x)**self.alpha)**exponent
        return f1*f2

    def _integ_interp(self):
        # Precomputes the Fourier transform of the profile in terms
        # of the scaled radius x and creates a spline interpolator
        # for it.
        from scipy.interpolate import interp1d
        from scipy.integrate import quad

        def integrand(x):
            return self._form_factor(x)*x

        q_arr = np.geomspace(self.qrange[0], self.qrange[1], self.nq)
        # We use the `weight` feature of quad to quickly estimate
        # the Fourier transform. We could use the existing FFTLog
        # framework, but this is a lot less of a kerfuffle.
        f_arr = np.array([quad(integrand,
                               a=1e-4, b=self.x_out,  # limits of integration
                               weight="sin",  # fourier sine weight
                               wvar=q)[0] / q
                          for q in q_arr])
        Fq = interp1d(np.log(q_arr), f_arr,
                      fill_value="extrapolate",
                      bounds_error=False)
        return Fq

    def _norm(self, cosmo, M, a, mb):
        # Computes the normalization factor of the GNFW profile.
        # Normalization factor is given in units of eV/cm^3.
        # (Bolliet et al. 2017).
        h70 = cosmo["h"]/0.7
        C0 = 1.65*h70**2
        CM = (h70*M*mb/3E14)**(2/3+self.alpha_P)   # M dependence
        Cz = cosmo.h_over_h0(a)**(8/3)  # z dependence
        P0_corr = self.P0 * h70**self.P0_hexp  # h-corrected P_0
        return P0_corr * C0 * CM * Cz

    def _real(self, cosmo, r, M, a):
        # Real-space profile.
        # Output in units of eV/cm^3
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        # (1-b)
        mb = self.mass_bias
        # R_Delta*(1+z)
        R = self.mass_def.get_radius(cosmo, M_use * mb, a) / a

        nn = self._norm(cosmo, M_use, a, mb)
        prof = self._form_factor(r_use[None, :] / R[:, None])
        prof *= nn[:, None]

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fourier(self, cosmo, k, M, a):
        # Fourier-space profile.
        # Output in units of eV * Mpc^3 / cm^3.

        # Tabulate if not done yet
        if self._fourier_interp is None:
            with UnlockInstance(self):
                self._fourier_interp = self._integ_interp()

        # Input handling
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        mb = self.mass_bias
        # R_Delta*(1+z)
        R = self.mass_def.get_radius(cosmo, M_use*mb, a) / a

        ff = self._fourier_interp(np.log(k_use[None, :] * R[:, None]))
        nn = self._norm(cosmo, M_use, a, mb)

        prof = (4*np.pi*R**3 * nn)[:, None] * ff

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof



__all__ = ("HaloProfileEinasto",)

import numpy as np
from scipy.integrate import quad_vec
from scipy.special import gamma, gammainc

from .. import MassDef, mass_translator, get_delta_c
from . import HaloProfileMatter


class HaloProfileEinasto(HaloProfileMatter):
    """ `Einasto 1965
    <https://ui.adsabs.harvard.edu/abs/1965TrAlm...5...87E/abstract>`_
    profile.

    .. math::
       \\rho(r) = \\rho_0\\,\\exp(-2 ((r/r_s)^\\alpha-1) / \\alpha)

    where :math:`r_s` is related to the comoving spherical overdensity
    halo radius :math:`r_\\Delta(M)` through the concentration
    parameter :math:`c(M)` as

    .. math::
       r_\\Delta(M) = c(M)\\,r_s

    and the normalization :math:`\\rho_0` is

    .. math::
       \\rho_0 = \\frac{M}{4\\pi\\,r_s^3}
       \\frac{2^{(3/\\alpha)}\\,\\alpha^{(1-3/\\alpha)}
       \\,{\\rm exp}(-2/\\alpha)}
       {\\gamma(\\frac{3}{\\alpha}, \\frac{2}{\\alpha}c^{\\alpha})},

    where :math:`\\gamma` is the lower incomplete gamma function.

    The index :math:`\\alpha` can be a free parameter or dependent on halo mass
    and redshift. In the latter case, we use the parameterization of
    `Diemer & Kravtsov <https://arxiv.org/abs/1401.1216>`_.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        concentration (:class:`~pyccl.halos.halo_model_base.Concentration`):
            concentration-mass relation to use with this profile.
        truncated (:obj:`bool`): set to ``True`` if the profile should be
            truncated at :math:`r = r_\\Delta`.
        projected_quad (:obj:`bool`): set to ``True`` to calculate the
            projected profile with numerical integration.
        alpha (:obj:`float` or :obj:`str`): :math:`\\alpha` parameter, or
            set to ``'cosmo'`` to calculate the value from cosmology.
    """
    __repr_attrs__ = __eq_attrs__ = (
        "truncated", "alpha", "projected_quad",
        "mass_def", "concentration", "precision_fftlog",)

    def __init__(self, *, mass_def, concentration,
                 truncated=False,
                 projected_quad=False,
                 alpha='cosmo'):
        self.truncated = truncated
        self.projected_quad = projected_quad
        self.alpha = alpha
        if projected_quad:
            if truncated:
                raise ValueError("projected_quad profile not supported "
                                 "for truncated Einasto. Set `truncated` or "
                                 "`projected_quad` to `False`.")
            self._projected = self._projected_quad
        super().__init__(mass_def=mass_def, concentration=concentration)
        self._to_virial_mass = mass_translator(
            mass_in=self.mass_def, mass_out=MassDef("vir", "matter"),
            concentration=self.concentration)
        self.update_precision_fftlog(padding_hi_fftlog=1E2,
                                     padding_lo_fftlog=1E-2,
                                     n_per_decade=1000,
                                     plaw_fourier=-2.)

    def update_parameters(self, alpha=None):
        """Update any of the parameters associated with this profile.
        Any parameter set to ``None`` won't be updated.

        Args:
            alpha (:obj:`float` or :obj:`str`): :math:`\\alpha` parameter, or
                set to ``'cosmo'`` to calculate the value from cosmology.
        """
        if alpha is not None and alpha != self.alpha:
            self.alpha = alpha

    def _get_alpha(self, cosmo, M, a):
        if self.alpha == 'cosmo':
            Mvir = self._to_virial_mass(cosmo, M, a)
            sM = cosmo.sigmaM(Mvir, a)
            nu = get_delta_c(cosmo, a, kind='EdS_approx') / sM
            return 0.155 + 0.0095 * nu * nu
        return np.full_like(M, self.alpha)

    def _norm(self, M, Rs, c, alpha):
        # Einasto normalization from mass, radius, concentration and alpha
        return M / (np.pi * Rs**3 * 2**(2-3/alpha) * alpha**(-1+3/alpha)
                    * np.exp(2/alpha)
                    * gamma(3/alpha) * gammainc(3/alpha, 2/alpha*c**alpha))

    def _real(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        alpha = self._get_alpha(cosmo, M_use, a)

        norm = self._norm(M_use, R_s, c_M, alpha)

        x = r_use[None, :] / R_s[:, None]
        prof = norm[:, None] * np.exp(-2. * (x**alpha[:, None] - 1) /
                                      alpha[:, None])
        if self.truncated:
            prof[r_use[None, :] > R_M[:, None]] = 0

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _projected_quad(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        alpha = self._get_alpha(cosmo, M_use, a)

        prof, _ = quad_vec(
            self._projected_quad_integrand, 0., np.inf,
            args=(r_use[None, :], R_s[:, None], alpha[:, None]))

        prof *= 2 * self._norm(M_use, R_s, c_M, alpha)[:, None]

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _projected_quad_integrand(self, z, R, R_s, alpha):
        x = np.sqrt(z**2. + R**2.) / R_s
        return np.exp(-2. * (x**alpha - 1.) / alpha)



import numpy as np
import pytest
import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG


POWER_MG_TOL = 1e-4


@pytest.mark.parametrize('model', list(range(5)))
def test_power_mg(model):
    mu_0 = [0., 0.1, -0.1, 0.1, -0.1]
    sigma_0 = [0., 0.1, -0.1, -0.1, 0.1]
    h0 = 0.7
    cosmoMG = ccl.Cosmology(
        Omega_c=0.112/h0**2,
        Omega_b=0.0226/h0**2,
        h=h0,
        A_s=2.1e-9,
        n_s=0.96,
        Neff=3.046,
        Omega_k=0,
        m_nu=0,
        T_CMB=2.7255,
        mg_parametrization=MuSigmaMG(
            mu_0=mu_0[model],
            sigma_0=sigma_0[model]),
        matter_power_spectrum='linear',
        transfer_function='boltzmann_isitgr')

    data = np.loadtxt("./benchmarks/data/model%d_pk_MG_matterpower.dat"
                      % model)

    a = 1
    k = data[:, 0] * cosmoMG['h']
    pk = data[:, 1] / (cosmoMG['h']**3)
    pk_ccl = ccl.linear_matter_power(cosmoMG, k, a)
    err = np.abs(pk_ccl/pk - 1)
    print(cosmoMG)
    # cut two points due to cosmic variance
    cut = data[:, 0] > 1e-04
    assert np.allclose(err[cut], 0, rtol=0, atol=POWER_MG_TOL)



import numpy as np
import pytest
import pyccl as ccl


@pytest.mark.parametrize('split', ['normal', 'inverted', 'equal', 'sum',
                                   'single'])
def test_nu_masses_smoke(split):
    m = ccl.nu_masses(Omega_nu_h2=0.1, mass_split=split)
    if split == "sum":
        assert np.shape(m) == ()
    elif split == "single":
        assert np.shape(m) == (1,)
    else:
        assert np.shape(m) == (3,)


def test_neutrinos_raises():
    with pytest.raises(ValueError):
        ccl.nu_masses(Omega_nu_h2=0.1, mass_split='blah')



__all__ = ("CCLError", "CCLWarning", "CCLDeprecationWarning",)

import warnings


class CCLError(RuntimeError):
    """A CCL-specific RuntimeError"""
    def __repr__(self):
        return "pyccl.CCLError(%r)" % (str(self))

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))


class CCLWarning(RuntimeWarning):
    """A CCL-specific warning"""
    def __repr__(self):
        return "pyccl.CCLWarning(%r)" % (str(self))

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))


class CCLDeprecationWarning(DeprecationWarning):
    """A CCL-specific deprecation warning."""
    def __repr__(self):
        return "pyccl.CCLDeprecationWarning(%r)" % (str(self))

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))

    @classmethod
    def enable(cls):
        warnings.simplefilter("always")

    @classmethod
    def disable(cls):
        warnings.filterwarnings(action="ignore", category=cls)



__all__ = ("EulerianPTCalculator",)

import warnings

import numpy as np

from .. import (CCLAutoRepr, CCLError, CCLWarning, Pk2D,
                get_pk_spline_a, unlock_instance)


# All valid Pk pair labels and their aliases
_PK_ALIAS = {
    'm:m': 'm:m', 'm:b1': 'm:m', 'm:b2': 'm:b2',
    'm:b3nl': 'm:b3nl', 'm:bs': 'm:bs', 'm:bk2': 'm:bk2',
    'm:c1': 'm:m', 'm:c2': 'm:c2', 'm:cdelta': 'm:cdelta',
    'b1:b1': 'm:m', 'b1:b2': 'm:b2', 'b1:b3nl': 'm:b3nl',
    'b1:bs': 'm:bs', 'b1:bk2': 'm:bk2', 'b1:c1': 'm:m',
    'b1:c2': 'm:c2', 'b1:cdelta': 'm:cdelta', 'b2:b2': 'b2:b2',
    'b2:b3nl': 'zero', 'b2:bs': 'b2:bs', 'b2:bk2': 'zero',
    'b2:c1': 'zero', 'b2:c2': 'zero', 'b2:cdelta': 'zero',
    'b3nl:b3nl': 'zero', 'b3nl:bs': 'zero',
    'b3nl:bk2': 'zero', 'b3nl:c1': 'zero', 'b3nl:c2':
    'zero', 'b3nl:cdelta': 'zero', 'bs:bs': 'bs:bs',
    'bs:bk2': 'zero', 'bs:c1': 'zero', 'bs:c2': 'zero',
    'bs:cdelta': 'zero', 'bk2:bk2': 'zero', 'bk2:c1': 'zero',
    'bk2:c2': 'zero', 'bk2:cdelta': 'zero', 'c1:c1': 'm:m',
    'c1:c2': 'm:c2', 'c1:cdelta': 'm:cdelta', 'c2:c2': 'c2:c2',
    'c2:cdelta': 'c2:cdelta', 'cdelta:cdelta': 'cdelta:cdelta'}


class EulerianPTCalculator(CCLAutoRepr):
    """ This class implements a set of methods that can be
    used to compute the various components needed to estimate
    Eulerian perturbation theory correlations. These calculations
    are currently based on FAST-PT
    (https://github.com/JoeMcEwen/FAST-PT).

    In the parametrisation used here, the galaxy overdensity
    is expanded as:

    .. math::
        \\delta_g=b_1\\,\\delta+\\frac{b_2}{2}\\delta^2+
        \\frac{b_s}{2}s^2+\\frac{b_{3nl}}{2}\\psi_{nl}+
        \\frac{b_{k2}}{2}\\nabla^2\\delta.

    In turn, the intrinsic alignment component is expanded as

    .. math::
        s^I_{ij}=c_1\\,s_{ij}+c_2(s_{ik}s_{jk}-s^2\\delta_{ik}/3)
        +c_\\delta\\,\\delta\\,s_{ij}

    (note that the higher-order terms are not divided by 2!).

    .. note:: Only the leading-order non-local term (i.e.
              :math:`\\langle \\delta\\,\\nabla^2\\delta`) is
              taken into account in the expansion. All others are
              set to zero.

    .. note:: Terms of the form
              :math:`\\langle \\delta^2 \\psi_{nl}\\rangle` (and
              likewise for :math:`s^2`) are set to zero.

    .. warning:: The full non-linear model for the cross-correlation
                 between number counts and intrinsic alignments is
                 still work in progress in FastPT. As a workaround
                 CCL assumes a non-linear treatment of IAs, but only
                 linearly biased number counts.

    .. note:: This calculator does not account for any form of
              stochastic bias contribution to the power spectra.
              If necessary, consider adding it in post-processing.

    Args:
        with_NC (:obj:`bool`): set to ``True`` if you'll want to use
            this calculator to compute correlations involving
            number counts.
        with_IA(:obj:`bool`): set to ``True`` if you'll want to use
            this calculator to compute correlations involving
            intrinsic alignments.
        with_matter_1loop(:obj:`bool`): set to ``True`` if you'll want to use
            this calculator to compute the one-loop matter power
            spectrum (automatically on if ``with_NC==True``).
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            If present, internal PT power spectrum templates will
            be initialized. If ``None``, you will need to initialize
            them using the :meth:`update_ingredients` method.
        log10k_min (:obj:`float`): decimal logarithm of the minimum
            Fourier scale (in :math:`{\\rm Mpc}^{-1}`) for which you want to
            calculate perturbation theory quantities.
        log10k_max (:obj:`float`): decimal logarithm of the maximum
            Fourier scale (in :math:`{\\rm Mpc}^{-1}`) for which you want to
            calculate perturbation theory quantities.
        nk_per_decade (:obj:`int` or :obj:`float`): number of k values per
            decade.
        a_arr (array): array of values of the scale factor at
            which all power spectra will be evaluated. If ``None``,
            the default sampling used internally by CCL will be
            used. Note that this may be slower than a bespoke sampling
            optimised for your particular application.
        k_cutoff (:obj:`float`): exponential cutoff scale. All power
            spectra will be multiplied by a cutoff factor of the
            form :math:`\\exp(-(k/k_*)^n)`, where :math:`k_*` is
            the cutoff scale. This may be useful when using the
            resulting power spectra to compute correlation
            functions if some of the PT contributions do not
            fall sufficiently fast on small scales. If ``None``
            (default), no cutoff factor will be applied.
        n_exp_cutoff (:obj:`float`): exponent of the cutoff factor (see
            ``k_cutoff``).
        b1_pk_kind (:obj:`str`): power spectrum to use for the first-order
            bias terms in the expansion. ``'linear'``: use the linear
            matter power spectrum. ``'nonlinear'``: use the non-linear
            matter power spectrum. ``'pt'``: use the 1-loop SPT matter
            power spectrum.
        bk2_pk_kind (:obj:`str`): power spectrum to use for the non-local
            bias terms in the expansion. Same options and default as
            ``b1_pk_kind``.
        pad_factor (:obj:`float`): fraction of the :math:`\\log_{10}(k)`
             interval you to add as padding for FFTLog calculations.
        low_extrap (:obj:`float`): decimal logaritm of the minimum Fourier
             scale (in :math:`{\\rm Mpc}^{-1}`) for which FAST-PT will
             extrapolate.
        high_extrap (:obj:`float`): decimal logaritm of the maximum Fourier
             scale (in :math:`{\\rm Mpc}^{-1}`) for which FAST-PT will
             extrapolate.
        P_window (array): 2-element array describing the
             tapering window used by FAST-PT. See FAST-PT
             documentation for more details.
        C_window (:obj:`float`):  `C_window` parameter used by FAST-PT to
             smooth the edges and avoid ringing. See FAST-PT
             documentation for more details.
        sub_lowk (:obj:`bool`): if ``True``, the small-scale white noise
             contribution to some of the terms will be subtracted.
    """
    __repr_attrs__ = __eq_attrs__ = ('with_NC', 'with_IA', 'with_matter_1loop',
                                     'k_s', 'a_s', 'exp_cutoff', 'b1_pk_kind',
                                     'bk2_pk_kind', 'fastpt_par', )

    def __init__(self, *, with_NC=False, with_IA=False,
                 with_matter_1loop=True, cosmo=None,
                 log10k_min=-4, log10k_max=2, nk_per_decade=20,
                 a_arr=None, k_cutoff=None, n_exp_cutoff=4,
                 b1_pk_kind='nonlinear', bk2_pk_kind='nonlinear',
                 pad_factor=1.0, low_extrap=-5.0, high_extrap=3.0,
                 P_window=None, C_window=0.75, sub_lowk=False):
        self.with_matter_1loop = with_matter_1loop
        self.with_NC = with_NC
        self.with_IA = with_IA

        # Set FAST-PT parameters
        self.fastpt_par = {'pad_factor': pad_factor,
                           'low_extrap': low_extrap,
                           'high_extrap': high_extrap,
                           'P_window': P_window,
                           'C_window': C_window,
                           'sub_lowk': sub_lowk}

        to_do = ['one_loop_dd']
        if self.with_NC:
            to_do.append('dd_bias')
        if self.with_IA:
            to_do.append('IA')

        # k sampling
        nk_total = int((log10k_max - log10k_min) * nk_per_decade)
        self.k_s = np.logspace(log10k_min, log10k_max, nk_total)

        # a sampling
        if a_arr is None:
            a_arr = get_pk_spline_a()
        self.a_s = a_arr.copy()
        self.z_s = 1/self.a_s-1

        # Cutoff factor
        if k_cutoff is not None:
            self.exp_cutoff = np.exp(-(self.k_s/k_cutoff)**n_exp_cutoff)
            self.exp_cutoff = self.exp_cutoff[None, :]
        else:
            self.exp_cutoff = 1

        # Call FAST-PT
        import fastpt as fpt
        n_pad = int(self.fastpt_par['pad_factor'] * len(self.k_s))
        self.pt = fpt.FASTPT(self.k_s, to_do=to_do,
                             low_extrap=self.fastpt_par['low_extrap'],
                             high_extrap=self.fastpt_par['high_extrap'],
                             n_pad=n_pad)

        # b1/bk P(k) prescription
        if b1_pk_kind not in ['linear', 'nonlinear', 'pt']:
            raise ValueError(f"Unknown P(k) prescription {b1_pk_kind}")
        if bk2_pk_kind not in ['linear', 'nonlinear', 'pt']:
            raise ValueError(f"Unknown P(k) prescription {bk2_pk_kind}")
        self.b1_pk_kind = b1_pk_kind
        self.bk2_pk_kind = bk2_pk_kind
        if (self.b1_pk_kind == 'pt') or (self.bk2_pk_kind == 'pt'):
            self.with_matter_1loop = True

        # Initialize all expensive arrays to ``None``.
        self._cosmo = None

        # Fill them out if cosmo is present
        if cosmo is not None:
            self.update_ingredients(cosmo)

        # All valid Pk pair labels
        self._pk_valid = list(_PK_ALIAS.keys())
        # List of Pk2Ds to fill out
        self._pk2d_temp = {}

    def _check_init(self):
        if self.initialised:
            return
        raise CCLError("PT templates have not been initialised "
                       "for this calculator. Please do so using "
                       "`update_ingredients`.")

    @property
    def initialised(self):
        return hasattr(self, "pk_bk")

    @unlock_instance
    def update_ingredients(self, cosmo):
        """ Update the internal PT arrays.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        """
        if self.initialised and (cosmo == self._cosmo):
            return

        pklz0 = cosmo.linear_matter_power(self.k_s, 1.0)
        g = cosmo.growth_factor(self.a_s)
        self._g4 = g**4
        self._g4T = self._g4[:, None]

        kw = {'P': pklz0, 'P_window': self.fastpt_par['P_window'],
              'C_window': self.fastpt_par['C_window']}

        def reshape_fastpt(tupl):
            for qq in tupl:
                if np.ndim(qq) > 0:
                    qq = qq[None, :]

        # Galaxy clustering templates
        if self.with_NC:
            self.dd_bias = self.pt.one_loop_dd_bias_b3nl(**kw)
            reshape_fastpt(self.dd_bias)
            self.one_loop_dd = self.dd_bias[0:1]
            self.with_matter_1loop = True
        elif self.with_matter_1loop:  # Only 1-loop matter needed
            self.one_loop_dd = self.pt.one_loop_dd(**kw)
            reshape_fastpt(self.one_loop_dd)

        # Intrinsic alignment templates
        if self.with_IA:
            self.ia_ta = self.pt.IA_ta(**kw)
            reshape_fastpt(self.ia_ta)
            self.ia_tt = self.pt.IA_tt(**kw)
            reshape_fastpt(self.ia_tt)
            self.ia_mix = self.pt.IA_mix(**kw)
            reshape_fastpt(self.ia_mix)

        # b1/bk power spectrum
        pks = {}
        if 'nonlinear' in [self.b1_pk_kind, self.bk2_pk_kind]:
            pks['nonlinear'] = np.array([cosmo.nonlin_matter_power(self.k_s, a)
                                         for a in self.a_s])
        if 'linear' in [self.b1_pk_kind, self.bk2_pk_kind]:
            pks['linear'] = np.array([cosmo.linear_matter_power(self.k_s, a)
                                      for a in self.a_s])
        if 'pt' in [self.b1_pk_kind, self.bk2_pk_kind]:
            if 'linear' in pks:
                pk = pks['linear']
            else:
                pk = np.array([cosmo.linear_matter_power(self.k_s, a)
                               for a in self.a_s])
            # Add SPT correction
            pk += self._g4T * self.one_loop_dd[0]
            pks['pt'] = pk
        self.pk_b1 = pks[self.b1_pk_kind]
        self.pk_bk = pks[self.bk2_pk_kind]

        # Reset template power spectra
        self._pk2d_temp = {}
        self._cosmo = cosmo

    def _get_pgg(self, tr1, tr2):
        """ Get the number counts auto-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            tr1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first
                tracer to correlate.
            tr2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first
                tracer to correlate.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get Pk templates
        Pd1d1 = self.pk_b1
        Pd1d2 = self._g4T * self.dd_bias[2]
        Pd2d2 = self._g4T * self.dd_bias[3]
        Pd1s2 = self._g4T * self.dd_bias[4]
        Pd2s2 = self._g4T * self.dd_bias[5]
        Ps2s2 = self._g4T * self.dd_bias[6]
        Pd1p3 = self._g4T * self.dd_bias[8]
        Pd1k2 = self.pk_bk * (self.k_s**2)[None, :]

        # Get biases
        b11 = tr1.b1(self.z_s)
        b21 = tr1.b2(self.z_s)
        bs1 = tr1.bs(self.z_s)
        bk21 = tr1.bk2(self.z_s)
        b3nl1 = tr1.b3nl(self.z_s)
        b12 = tr2.b1(self.z_s)
        b22 = tr2.b2(self.z_s)
        bs2 = tr2.bs(self.z_s)
        bk22 = tr2.bk2(self.z_s)
        b3nl2 = tr2.b3nl(self.z_s)

        s4 = 0.
        if self.fastpt_par['sub_lowk']:
            s4 = self._g4T * self.dd_bias[7]

        pgg = ((b11*b12)[:, None] * Pd1d1 +
               0.5*(b11*b22 + b12*b21)[:, None] * Pd1d2 +
               0.25*(b21*b22)[:, None] * (Pd2d2 - 2.*s4) +
               0.5*(b11*bs2 + b12*bs1)[:, None] * Pd1s2 +
               0.25*(b21*bs2 + b22*bs1)[:, None] * (Pd2s2 - (4./3.)*s4) +
               0.25*(bs1*bs2)[:, None] * (Ps2s2 - (8./9.)*s4) +
               0.5*(b12*b3nl1+b11*b3nl2)[:, None] * Pd1p3 +
               0.5*(b12*bk21+b11*bk22)[:, None] * Pd1k2)

        return pgg*self.exp_cutoff

    def _get_pgi(self, trg, tri):
        """ Get the number counts - IA cross-spectrum at the internal
        set of wavenumbers and scale factors.

        .. note:: The full non-linear model for the cross-correlation
                  between number counts and intrinsic alignments is
                  still work in progress in FastPT. As a workaround
                  CCL assumes a non-linear treatment of IAs, but only
                  linearly biased number counts.

        Args:
            trg (:class:`~pyccl.nl_pt.tracers.PTTracer`): number
                counts tracer.
            tri (:class:`~pyccl.nl_pt.tracers.PTTracer`): intrinsic
                alignment tracer.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get Pk templates
        Pd1d1 = self.pk_b1
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        # Get biases
        b1 = trg.b1(self.z_s)
        b2 = trg.b2(self.z_s)
        bs = trg.bs(self.z_s)
        bk2 = trg.bk2(self.z_s)
        b3nl = trg.b3nl(self.z_s)
        if any([b.any() for b in [b2, bs, bk2, b3nl]]):
            warnings.warn(
                "EulerianPTCalculators assume linear galaxy bias "
                "when computing galaxy-IA cross-correlations.",
                category=CCLWarning)
        c1 = tri.c1(self.z_s)
        c2 = tri.c2(self.z_s)
        cd = tri.cdelta(self.z_s)

        pgi = b1[:, None] * (c1[:, None] * Pd1d1 +
                             (self._g4*cd)[:, None] * (a00e + c00e) +
                             (self._g4*c2)[:, None] * (a0e2 + b0e2))
        return pgi*self.exp_cutoff

    def _get_pgm(self, trg):
        """ Get the number counts - matter cross-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            trg (:class:`~pyccl.nl_pt.tracers.PTTracer`): number
                counts tracer.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get Pk templates
        Pd1d1 = self.pk_b1
        Pd1d2 = self._g4T * self.dd_bias[2]
        Pd1s2 = self._g4T * self.dd_bias[4]
        Pd1p3 = self._g4T * self.dd_bias[8]
        Pd1k2 = self.pk_bk*(self.k_s**2)[None, :]

        # Get biases
        b1 = trg.b1(self.z_s)
        b2 = trg.b2(self.z_s)
        bs = trg.bs(self.z_s)
        bk2 = trg.bk2(self.z_s)
        b3nl = trg.b3nl(self.z_s)

        pgm = (b1[:, None] * Pd1d1 +
               0.5 * b2[:, None] * Pd1d2 +
               0.5 * bs[:, None] * Pd1s2 +
               0.5 * b3nl[:, None] * Pd1p3 +
               0.5 * bk2[:, None] * Pd1k2)

        return pgm*self.exp_cutoff

    def _get_pii(self, tr1, tr2, return_bb=False):
        """ Get the intrinsic alignment auto-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            tr1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first tracer
                to correlate.
            tr2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first tracer
                to correlate.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get Pk templates
        Pd1d1 = self.pk_b1
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        ae2e2, ab2b2 = self.ia_tt
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        # Get biases
        c11 = tr1.c1(self.z_s)
        c21 = tr1.c2(self.z_s)
        cd1 = tr1.cdelta(self.z_s)
        c12 = tr2.c1(self.z_s)
        c22 = tr2.c2(self.z_s)
        cd2 = tr2.cdelta(self.z_s)

        if return_bb:
            pii = ((cd1*cd2*self._g4)[:, None]*a0b0b +
                   (c21*c22*self._g4)[:, None]*ab2b2 +
                   ((cd1*c22+c21*cd2)*self._g4)[:, None] * d0bb2)
        else:
            pii = ((c11*c12)[:, None] * Pd1d1 +
                   ((c11*cd2+c12*cd1)*self._g4)[:, None]*(a00e+c00e) +
                   (cd1*cd2*self._g4)[:, None]*a0e0e +
                   (c21*c22*self._g4)[:, None]*ae2e2 +
                   ((c11*c22+c21*c12)*self._g4)[:, None]*(a0e2+b0e2) +
                   ((cd1*c22+cd2*c21)*self._g4)[:, None]*d0ee2)

        return pii*self.exp_cutoff

    def _get_pim(self, tri):
        """ Get the matter - IA cross-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            tri (:class:`~pyccl.nl_pt.tracers.PTTracer`): intrinsic
                alignment tracer.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get Pk templates
        Pd1d1 = self.pk_b1
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        # Get biases
        c1 = tri.c1(self.z_s)
        c2 = tri.c2(self.z_s)
        cd = tri.cdelta(self.z_s)

        pim = (c1[:, None] * Pd1d1 +
               (self._g4*cd)[:, None] * (a00e + c00e) +
               (self._g4*c2)[:, None] * (a0e2 + b0e2))
        return pim*self.exp_cutoff

    def _get_pmm(self):
        """ Get the one-loop matter power spectrum.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        if self.b1_pk_kind == 'linear':
            P1loop = self._g4T * self.one_loop_dd[0]
        else:
            P1loop = 0.
        pk = self.pk_b1 + P1loop
        return pk*self.exp_cutoff

    def get_biased_pk2d(self, tracer1, *, tracer2=None, return_ia_bb=False,
                        extrap_order_lok=1, extrap_order_hik=2):
        """Returns a :class:`~pyccl.pk2d.Pk2D` object containing
        the PT power spectrum for two quantities defined by
        two :class:`~pyccl.nl_pt.tracers.PTTracer` objects.

        .. note:: The full non-linear model for the cross-correlation
                  between number counts and intrinsic alignments is
                  still work in progress in FastPT. As a workaround
                  CCL assumes a non-linear treatment of IAs, but only
                  linearly biased number counts.

        Args:
            tracer1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the first
                tracer being correlated.
            tracer2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the second
                tracer being correlated. If ``None``, the auto-correlation
                of the first tracer will be returned.
            return_ia_bb (:obj:`bool`): if ``True``, the B-mode power spectrum
                for intrinsic alignments will be returned (if both
                input tracers are of type
                :class:`~pyccl.nl_pt.tracers.PTIntrinsicAlignmentTracer`)
                If ``False`` (default) E-mode power spectrum is returned.
            extrap_order_lok (:obj:`int`): extrapolation order to be used on
                k-values below the minimum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.
            extrap_order_hik (:obj:`int`): extrapolation order to be used on
                k-values above the maximum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.

        Returns:
            :class:`~pyccl.pk2d.Pk2D`: PT power spectrum.
        """
        if return_ia_bb:
            return_ia_bb = True

        if tracer2 is None:
            tracer2 = tracer1

        t1 = tracer1.type
        t2 = tracer2.type

        if ((t1 == 'NC') or (t2 == 'NC')) and (not self.with_NC):
            raise ValueError("Can't use number counts tracer in "
                             "EulerianPTCalculator with 'with_NC=False'")
        if ((t1 == 'IA') or (t2 == 'IA')) and (not self.with_IA):
            raise ValueError("Can't use intrinsic alignment tracer in "
                             "EulerianPTCalculator with 'with_IA=False'")

        if t1 == 'NC':
            if t2 == 'NC':
                pk = self._get_pgg(tracer1, tracer2)
            elif t2 == 'IA':
                pk = self._get_pgi(tracer1, tracer2)
            else:  # Must be matter
                pk = self._get_pgm(tracer1)
        elif t1 == 'IA':
            if t2 == 'NC':
                pk = self._get_pgi(tracer2, tracer1)
            elif t2 == 'IA':
                pk = self._get_pii(tracer1, tracer2,
                                   return_bb=return_ia_bb)
            else:  # Must be matter
                pk = self._get_pim(tracer1)
        else:  # Must be matter
            if t2 == 'NC':
                pk = self._get_pgm(tracer2)
            elif t2 == 'IA':
                pk = self._get_pim(tracer2)
            else:  # Must be matter
                pk = self._get_pmm()

        pk2d = Pk2D(a_arr=self.a_s,
                    lk_arr=np.log(self.k_s),
                    pk_arr=pk,
                    is_logp=False,
                    extrap_order_lok=extrap_order_lok,
                    extrap_order_hik=extrap_order_hik)
        return pk2d

    def get_pk2d_template(self, kind, *, extrap_order_lok=1,
                          extrap_order_hik=2, return_ia_bb=False):
        """Returns a :class:`~pyccl.pk2d.Pk2D` object containing
        the power spectrum template for two of the PT operators. The
        combination returned is determined by ``kind``, which must be
        a string of the form ``'q1:q2'``, where ``q1`` and ``q2`` denote
        the two operators whose power spectrum is sought. Valid
        operator names are: ``'m'`` (matter overdensity), ``'b1'``
        (first-order overdensity), ``'b2'`` (:math:`\\delta^2`
        term in galaxy bias expansion), ``'bs'`` (:math:`s^2` term
        in galaxy bias expansion), ``'b3nl'`` (:math:`\\psi_{nl}`
        term in galaxy bias expansion), ``'bk2'`` (non-local
        :math:`\\nabla^2 \\delta` term in galaxy bias expansion),
        ``'c1'`` (linear IA term), ``'c2'`` (:math:`s^2` term in IA
        expansion), ``'cdelta'`` (:math:`s\\delta` term in IA expansion).

        Args:
            kind (:obj:`str`): string defining the pair of PT operators for
                which we want the power spectrum.
            extrap_order_lok (:obj:`int`): extrapolation order to be used on
                k-values below the minimum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.
            extrap_order_hik (:obj:`int`): extrapolation order to be used on
                k-values above the maximum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.
            return_ia_bb (:obj:`bool`): if ``True``, the B-mode power spectrum
                for intrinsic alignments will be returned (if both
                input tracers are of type
                :class:`~pyccl.nl_pt.tracers.PTIntrinsicAlignmentTracer`)
                If ``False`` (default) E-mode power spectrum is returned.

        Returns:
            :class:`~pyccl.pk2d.Pk2D`: PT power spectrum.
        """
        if not (kind in _PK_ALIAS):
            # Reverse order and check again
            kind_reverse = ':'.join(kind.split(':')[::-1])
            if not (kind_reverse in _PK_ALIAS):
                raise ValueError(f"Pk template {kind} not valid")
            kind = kind_reverse
        pk_name = _PK_ALIAS[kind]

        if return_ia_bb and (pk_name in ['c2:c2', 'c2:cdelta',
                                         'cdelta:cdelta']):
            pk_name += '_bb'

        # If already built, return
        if pk_name in self._pk2d_temp:
            return self._pk2d_temp[pk_name]

        # Construct power spectrum array
        s4 = 0.
        if pk_name == 'm:m':
            pk = self.pk_b1
        elif pk_name == 'm:b2':
            pk = 0.5*self._g4T*self.dd_bias[2]
        elif pk_name == 'm:b3nl':
            pk = 0.5*self._g4T*self.dd_bias[8]
        elif pk_name == 'm:bs':
            pk = 0.5*self._g4T*self.dd_bias[4]
        elif pk_name == 'm:bk2':
            pk = 0.5*self.pk_bk*(self.k_s**2)
        elif pk_name == 'm:c2':
            pk = self._g4T * (self.ia_mix[0]+self.ia_mix[1])
        elif pk_name == 'm:cdelta':
            pk = self._g4T * (self.ia_ta[0]+self.ia_ta[1])
        elif pk_name == 'b2:b2':
            if self.fastpt_par['sub_lowk']:
                s4 = self.dd_bias[7]
            pk = 0.25*self._g4T*(self.dd_bias[3] - 2*s4)
        elif pk_name == 'b2:bs':
            if self.fastpt_par['sub_lowk']:
                s4 = self.dd_bias[7]
            pk = 0.25*self._g4T*(self.dd_bias[5] - 4*s4/3)
        elif pk_name == 'bs:bs':
            if self.fastpt_par['sub_lowk']:
                s4 = self.dd_bias[7]
            pk = 0.25*self._g4T*(self.dd_bias[6] - 8*s4/9)
        elif pk_name == 'c2:c2':
            pk = self._g4T * self.ia_tt[0]
        elif pk_name == 'c2:c2_bb':
            pk = self._g4T * self.ia_tt[1]
        elif pk_name == 'c2:cdelta':
            pk = self._g4T * self.ia_mix[2]
        elif pk_name == 'c2:cdelta_bb':
            pk = self._g4T * self.ia_mix[3]
        elif pk_name == 'cdelta:cdelta':
            pk = self._g4T * self.ia_ta[2]
        elif pk_name == 'cdelta:cdelta_bb':
            pk = self._g4T * self.ia_ta[3]
        elif pk_name == 'zero':
            # If zero, store None and return
            self._pk2d_temp[pk_name] = None
            return None

        # Build interpolator
        pk2d = Pk2D(a_arr=self.a_s,
                    lk_arr=np.log(self.k_s),
                    pk_arr=pk,
                    is_logp=False,
                    extrap_order_lok=extrap_order_lok,
                    extrap_order_hik=extrap_order_hik)

        # Store and return
        self._pk2d_temp[pk_name] = pk2d
        return pk2d



__all__ = ("HaloBiasSheth01",)

from . import HaloBias, get_delta_c


class HaloBiasSheth01(HaloBias):
    """Implements halo bias as described in `Sheth et al. 2001
    <https://arxiv.org/abs/astro-ph/9907024>`_. This
    parametrization is only valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = "Sheth01"

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _setup(self):
        self.a = 0.707
        self.sqrta = 0.84083292038
        self.b = 0.5
        self.c = 0.6
        self.dc = get_delta_c(None, None, kind='EdS')

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc/sigM
        anu2 = self.a * nu**2
        anu2c = anu2**self.c
        t1 = self.b * (1.0 - self.c) * (1.0 - 0.5 * self.c)
        return 1 + (self.sqrta * anu2 * (1 + self.b / anu2c) -
                    anu2c / (anu2c + t1)) / (self.sqrta * self.dc)



r"""
Tracers represent projected quantities that can be cross-correlated to get
angular power spectra (see :func:`~pyccl.cells.angular_cl`). In the most
general case, the angular power spectrum between two tracers is given by

.. math::
    C^{\alpha\beta}_\ell=\frac{2}{\pi}\int d\chi_1\,d\chi_2\,dk\,k^2
    P_{\alpha\beta}(k,\chi_1,\chi_2)\,\Delta^\alpha_\ell(k,\chi_1)\,
    \Delta^\beta_\ell(k,\chi_2),

where :math:`P_{\alpha\beta}` is a generalized power spectrum (see
:class:`~pyccl.pk2d.Pk2D`),
and :math:`\Delta^\alpha_\ell(k,\chi)` is a sum over different contributions
associated to tracer :math:`\alpha`, where every contribution takes the form:

.. math::
    \Delta^\alpha_\ell(k,\chi)=f^\alpha_\ell\,W_\alpha(\chi)\,
    T_\alpha(k,\chi)\,j^{(n_\alpha)}_\ell(k\chi).

Here, :math:`f^\alpha_\ell` is an :math:`\ell`-dependent **prefactor**,
:math:`W_\alpha(\chi)` is the **radial kernel**, :math:`T_\alpha(k,\chi)`
is the **transfer function**, and :math:`j^{(n)}_\ell(x)` is a
generalized version of the **spherical Bessel functions**. Descriptions
of each of these ingredients, and how to implement generalised Tracers
as sub-classes of the :class:`Tracer` base class can be found below. The
documentation of the base :class:`Tracer` class is a good place to start.
"""

import warnings

import numpy as np

from . import ccllib as lib
from .pyutils import check
from .errors import CCLWarning
from ._core.parameters import physical_constants
from ._core import CCLObject, UnlockInstance, unlock_instance
from .pyutils import (_check_array_params, NoneArr, _vectorize_fn6,
                      _get_spline1d_arrays, _get_spline2d_arrays)


__all__ = ("get_density_kernel", "get_lensing_kernel", "get_kappa_kernel",
           "Tracer", "NzTracer", "NumberCountsTracer", "WeakLensingTracer",
           "CMBLensingTracer", "tSZTracer", "CIBTracer", "ISWTracer",)


def _Sig_MG(cosmo, a, k):
    """Redshift-dependent modification to Poisson equation for massless
    particles under modified gravity.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
        k (:obj:`float` or `array`): Wavenumber for scale

    Returns:
        (:obj:`float` or `array`): Modification to Poisson equation under
            modified gravity at scale factor a.
            Sig_MG is assumed to be proportional to Omega_Lambda(z),
            see e.g. Abbott et al. 2018, 1810.02499, Eq. 9.
    """
    cosmo.compute_distances()
    return _vectorize_fn6(lib.Sig_MG, lib.Sig_MG_vec, cosmo, a, k)


def _check_background_spline_compatibility(cosmo, z):
    """Check that a redshift array lies within the support of the
    CCL background splines.
    """
    cosmo.compute_distances()
    a_bg, _ = _get_spline1d_arrays(cosmo.cosmo.data.chi)
    a = 1/(1+z)

    if a.min() < a_bg.min() or a.max() > a_bg.max():
        raise ValueError(
            "Tracer has wider redshift support than internal CCL splines. "
            f"Tracer: z=[{1/a.max()-1}, {1/a.min()-1}]. "
            f"Background splines: z=[{1/a_bg.max()-1}, {1/a_bg.min()-1}].")


def get_density_kernel(cosmo, *, dndz):
    """This convenience function returns the radial kernel for
    galaxy-clustering-like tracers. Given an unnormalized
    redshift distribution, it returns two arrays: :math:`\\chi`,
    :math:`W(\\chi)`, where :math:`\\chi` is an array of radial
    distances in units of Mpc and :math:`W(\\chi) = p(z)\\,H(z)`, where
    :math:`H(z)` is the expansion rate in units of :math:`{\\rm Mpc}^{-1}`,
    and :math:`p(z)` is the normalized redshift distribution.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): cosmology object
            used to transform redshifts into distances.
        dndz (:obj:`tuple`): A tuple of arrays ``(z, N(z))``
            giving the redshift distribution of the objects.
            The units are arbitrary; ``N(z)`` will be normalized
            to unity.
    """
    z_n, n = _check_array_params(dndz, 'dndz')
    _check_background_spline_compatibility(cosmo, dndz[0])
    # this call inits the distance splines neded by the kernel functions
    chi = cosmo.comoving_radial_distance(1./(1.+z_n))
    status = 0
    wchi, status = lib.get_number_counts_kernel_wrapper(cosmo.cosmo,
                                                        z_n, n,
                                                        len(z_n),
                                                        status)
    check(status, cosmo=cosmo)
    return chi, wchi


def get_lensing_kernel(cosmo, *, dndz, mag_bias=None, n_chi=None):
    r"""This convenience function returns the radial kernel for
    weak-lensing-like. Given an unnormalized redshift distribution
    and an optional magnification bias function, it returns
    two arrays: :math:`\chi`, :math:`W(\chi)`, where :math:`\chi` is
    an array of radial distances in units of Mpc and :math:`W(\chi)` is
    the lensing kernel:

    .. math::
        W(\chi)=\frac{3H_0^2\Omega_m}{2a}\chi\,
        \int_{z(\chi)}^\infty dz\,\left(1-\frac{5s(z)}{2}\right)\,
        \frac{\chi(z)-\chi}{\chi(z)}

    .. note:: If using this function to compute the magnification bias
              kernel, the result must be multiplied by -2.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): cosmology object used to
            transform redshifts into distances.
        dndz (:obj:`tuple`): A tuple of arrays ``(z, N(z))``
            giving the redshift distribution of the objects.
            The units are arbitrary; ``N(z)`` will be normalized to unity.
        mag_bias (:obj:`tuple`): A tuple of arrays ``(z, s(z))``
            giving the magnification bias as a function of redshift. If
            ``None``, ``s=0`` will be assumed.
    """
    # we need the distance functions at the C layer
    cosmo.compute_distances()

    z_n, n = _check_array_params(dndz, 'dndz')
    has_magbias = mag_bias is not None
    z_s, s = _check_array_params(mag_bias, 'mag_bias')
    _check_background_spline_compatibility(cosmo, dndz[0])

    if n_chi is None:
        # Calculate number of samples in chi
        n_chi = lib.get_nchi_lensing_kernel_wrapper(z_n)

    if (n_chi > len(z_n)
            and cosmo.cosmo.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION):
        warnings.warn(
            f"The number of samples in the n(z) ({len(z_n)}) is smaller than "
            f"the number of samples in the lensing kernel ({n_chi}). Consider "
            "disabling spline integration for the lensing kernel by setting "
            "pyccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False "
            "before instantiating the Cosmology passed.", category=CCLWarning)

    # Compute array of chis
    status = 0
    chi, status = lib.get_chis_lensing_kernel_wrapper(cosmo.cosmo, z_n[-1],
                                                      n_chi, status)
    # Compute kernel
    wchi, status = lib.get_lensing_kernel_wrapper(cosmo.cosmo,
                                                  z_n, n, z_n[-1],
                                                  int(has_magbias), z_s, s,
                                                  chi, n_chi, status)
    check(status, cosmo=cosmo)
    return chi, wchi


def get_kappa_kernel(cosmo, *, z_source, n_samples=100):
    """This convenience function returns the radial kernel for
    CMB-lensing-like tracers.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        z_source (:obj:`float`): Redshift of source plane for CMB lensing.
        n_samples (:obj:`int`): number of samples over which the kernel
            is desired. These will be equi-spaced in radial distance.
            The kernel is quite smooth, so usually O(100) samples
            is enough.
    """
    _check_background_spline_compatibility(cosmo, np.array([z_source]))
    # this call inits the distance splines neded by the kernel functions
    chi_source = cosmo.comoving_radial_distance(1./(1.+z_source))
    chi = np.linspace(0, chi_source, n_samples)

    status = 0
    wchi, status = lib.get_kappa_kernel_wrapper(cosmo.cosmo, chi_source,
                                                chi, n_samples, status)
    check(status, cosmo=cosmo)
    return chi, wchi


class Tracer(CCLObject):
    """Tracers contain the information necessary to describe the
    contribution of a given sky observable to its cross-power spectrum
    with any other tracer. Tracers are composed of 4 main ingredients:

    * A radial kernel: this expresses the support in redshift/distance
      over which this tracer extends.

    * A transfer function: this is a function of wavenumber and
      scale factor that describes the connection between the tracer
      and the power spectrum on different scales and at different
      cosmic times.

    * An ell-dependent prefactor: normally associated with angular
      derivatives of a given fundamental quantity.

    * The order of the derivative of the Bessel functions with which
      they enter the computation of the angular power spectrum.

    A ``Tracer`` object will in reality be a list of different such
    tracers that get combined linearly when computing power spectra.
    """
    from ._core.repr_ import build_string_Tracer as __repr__

    def __init__(self):
        """By default this `Tracer` object will contain no actual
        tracers
        """
        # Do nothing, just initialize list of tracers
        self._trc = []

    def __eq__(self, other):
        # Check object id.
        if self is other:
            return True

        # Check the object class.
        if type(self) is not type(other):
            return False

        # If the tracer collections are empty, return early.
        if not (self or other):
            return True

        # If the tracer collections are not the same length, return early.
        if len(self._trc) != len(other._trc):
            return False

        # Check `der_angles` & `der_bessel` for each tracer in the collection.
        bessel = self.get_bessel_derivative(), other.get_bessel_derivative()
        angles = self.get_angles_derivative(), other.get_angles_derivative()
        if not (np.array_equal(*bessel) and np.array_equal(*angles)):
            return False

        # Check the kernels.
        for t1, t2 in zip(self._trc, other._trc):
            if bool(t1.kernel) ^ bool(t2.kernel):
                # only one of them has a kernel
                return False
            if t1.kernel is None:
                # none of them has a kernel
                continue
            if not np.array_equal(_get_spline1d_arrays(t1.kernel.spline),
                                  _get_spline1d_arrays(t2.kernel.spline)):
                # both have kernels, but they are unequal
                return False

        # Check the transfer functions.
        for t1, t2 in zip(self._trc, other._trc):
            if bool(t1.transfer) ^ bool(t2.transfer):
                # only one of them has a transfer
                return False
            if t1.transfer is None:
                # none of them has a transfer
                continue
            # Check the characteristics of the transfer function.
            for arg in ("extrap_order_lok", "extrap_order_hik",
                        "is_factorizable", "is_log"):
                if getattr(t1.transfer, arg) != getattr(t2.transfer, arg):
                    return False

            c2py = {"fa": _get_spline1d_arrays,
                    "fk": _get_spline1d_arrays,
                    "fka": _get_spline2d_arrays}
            for attr in c2py.keys():
                spl1 = getattr(t1.transfer, attr, None)
                spl2 = getattr(t2.transfer, attr, None)
                if bool(spl1) ^ bool(spl2):
                    # only one of them has this transfer type
                    return False
                if spl1 is None:
                    # none of them has this transfer type
                    continue
                # `pts` contain the the grid points and the transfer functions
                pts1, pts2 = c2py[attr](spl1), c2py[attr](spl2)
                for pt1, pt2 in zip(pts1, pts2):
                    # loop through output points of `_get_splinend_arrays`
                    if not np.array_equal(pt1, pt2):
                        # both have this transfer type, but they are unequal
                        # or are defined at different grid points
                        return False
        return True

    def __hash__(self):
        return hash(repr(self))

    def __bool__(self):
        return bool(self._trc)

    @property
    def chi_min(self):
        """Returns the minimum comoving distance over which this tracer's
        radial kernel is defined, if it exists. For tracers with more than
        one contribution with an associated ``chi_min``, the
        lowest value is returned.
        """
        chis = [tr.chi_min for tr in self._trc]
        return min(chis) if chis else None

    @property
    def chi_max(self):
        """Returns the maximum comoving distance over which this tracer's
        radial kernel is defined, if it exists. For tracers with more than
        one contribution with an associated ``chi_max``, the
        highest value is returned.
        """
        chis = [tr.chi_max for tr in self._trc]
        return max(chis) if chis else None

    def get_kernel(self, chi=None):
        """Get the radial kernels for all tracers contained
        in this ``Tracer``.

        Args:
            chi (:obj:`float` or `array`): values of the comoving
                radial distance in increasing order and in Mpc. If ``None``,
                returns the kernel at the internal spline nodes.

        Returns:
            `array`: list of radial kernels for each tracer. The shape
            will be ``(n_tracer, chi.size)``, where ``n_tracer`` is the
            number of tracers. The last dimension will be squeezed if the
            input is a scalar.
        """
        if chi is None:
            chis = []
        else:
            chi_use = np.atleast_1d(chi)

        kernels = []
        for t in self._trc:
            if t.kernel is None:
                continue
            else:
                if chi is None:
                    chi_use, w = _get_spline1d_arrays(t.kernel.spline)
                    chis.append(chi_use)
                else:
                    status = 0
                    w, status = lib.cl_tracer_get_kernel(
                        t, chi_use, chi_use.size, status)
                    check(status)
                kernels.append(w)

        if chi is None:
            return kernels, chis
        kernels = np.array(kernels)
        if np.ndim(chi) == 0 and kernels.shape != (0,):
            kernels = np.squeeze(kernels, axis=-1)
        return kernels

    def get_f_ell(self, ell):
        """Get the :math:`\\ell`-dependent prefactors for all tracers
        contained in this `Tracer`.

        Args:
            ell (:obj:`float` or `array`): angular multipole values.

        Returns:
            `array`: list of prefactors for each tracer.
            The shape will be ``(n_tracer, ell.size)``, where
            ``n_tracer`` is the number of tracers. The last
            dimension will be squeezed if the input ``ell`` is a
            scalar.
        """
        ell_use = np.atleast_1d(ell)
        f_ells = []
        for t in self._trc:
            status = 0
            f, status = lib.cl_tracer_get_f_ell(t, ell_use,
                                                ell_use.size,
                                                status)
            check(status)
            f_ells.append(f)
        f_ells = np.array(f_ells)
        if np.ndim(ell) == 0:
            if f_ells.shape != (0,):
                f_ells = np.squeeze(f_ells, axis=-1)
        return f_ells

    def get_transfer(self, lk, a):
        """Get the transfer functions for all tracers contained
        in this ``Tracer``.

        Args:
            lk (:obj:`float` or `array`): values of the natural logarithm of
                the wave number (in units of inverse Mpc) in increasing
                order.
            a (:obj:`float` or `array`): values of the scale factor.

        Returns:
            `array`: list of transfer functions for each tracer.
            The shape will be ``(n_tracer, lk.size, a.size)``, where
            ``n_tracer`` is the number of tracers. The other dimensions
            will be squeezed if the inputs are scalars.
        """
        lk_use = np.atleast_1d(lk)
        a_use = np.atleast_1d(a)
        transfers = []
        for t in self._trc:
            status = 0
            t, status = lib.cl_tracer_get_transfer(t, lk_use, a_use,
                                                   lk_use.size * a_use.size,
                                                   status)
            check(status)
            transfers.append(t.reshape([lk_use.size, a_use.size]))
        transfers = np.array(transfers)
        if transfers.shape != (0,):
            if np.ndim(a) == 0:
                transfers = np.squeeze(transfers, axis=-1)
                if np.ndim(lk) == 0:
                    transfers = np.squeeze(transfers, axis=-1)
            else:
                if np.ndim(lk) == 0:
                    transfers = np.squeeze(transfers, axis=-2)
        return transfers

    def get_bessel_derivative(self):
        """Get list of Bessel function derivative orders for all tracers
        contained in this ``Tracer``.

        Returns:
            `array`: list of Bessel derivative orders for each tracer.
        """
        return np.array([t.der_bessel for t in self._trc])

    def get_angles_derivative(self):
        r"""Get list of the :math:`\ell`-dependent prefactor order for all
        tracers contained in this ``Tracer``.

        Returns:
            `array`: list of angular derivative orders for each tracer.
        """
        return np.array([t.der_angles for t in self._trc])

    def _MG_add_tracer(self, cosmo, kernel, z_b, der_bessel=0, der_angles=0,
                       bias_transfer_a=None, bias_transfer_k=None):
        """ function to set mg_transfer in the right format and add MG tracers
            for different cases including different cases and biases like
            intrinsic alignements (IA) when present
        """
        # Getting MG transfer function and building a k-array
        mg_transfer = self._get_MG_transfer_function(cosmo, z_b)

        # case with no astro biases
        if ((bias_transfer_a is None) and (bias_transfer_k is None)):
            self.add_tracer(cosmo, kernel=kernel, transfer_ka=mg_transfer,
                            der_bessel=der_bessel, der_angles=der_angles)

        #  case of an astro bias depending on a and  k
        elif ((bias_transfer_a is not None) and (bias_transfer_k is not None)):
            mg_transfer_new = (mg_transfer[0], mg_transfer[1],
                               (bias_transfer_a[1] * (bias_transfer_k[1] *
                                mg_transfer[2]).T).T)
            self.add_tracer(cosmo, kernel=kernel, transfer_ka=mg_transfer_new,
                            der_bessel=der_bessel, der_angles=der_angles)

        #  case of an astro bias depending on a but not k
        elif ((bias_transfer_a is not None) and (bias_transfer_k is None)):
            mg_transfer_new = (mg_transfer[0], mg_transfer[1],
                               (bias_transfer_a[1] * mg_transfer[2].T).T)
            self.add_tracer(cosmo, kernel=kernel, transfer_ka=mg_transfer_new,
                            der_bessel=der_bessel, der_angles=der_angles)

        #  case of an astro bias depending on k but not a
        elif ((bias_transfer_a is None) and (bias_transfer_k is not None)):
            mg_transfer_new = (mg_transfer[0], mg_transfer[1],
                               (bias_transfer_k[1] * mg_transfer[2]))
            self.add_tracer(cosmo, kernel=kernel, transfer_ka=mg_transfer_new,
                            der_bessel=der_bessel, der_angles=der_angles)

    def _get_MG_transfer_function(self, cosmo, z):
        """ This function allows to obtain the function Sigma(z,k) (1 or 2D
            arrays) for an array of redshifts coming from a redshift
            distribution (defined by the user) and a single value or
            an array of k specified by the user. We obtain then Sigma(z,k) as a
            1D array for those z and k arrays and then convert it to a 2D array
            taking into consideration the given sizes of the arrays for z and k
            The MG parameter array goes then as a multiplicative factor within
            the MG transfer function. If k is not specified then only a 1D
            array for Sigma(a,k=0) is used.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): cosmology object used
                to transform redshifts into distances.
            z (float or tuple of arrays): a single z value (e.g. for CMB)
                or a tuple of arrays (z, N(z)) giving the redshift distribution
                of the objects. The units are arbitrary; N(z) will be
                normalized to unity.
            k (float or array): a single k value or an array of k for which we
                calculate the MG parameter Sigma(a,k). For now, the k range
                should be limited to linear scales.
        """
        # Sampling scale factor from a very small (at CMB for example)
        # all the way to 1 here and today for the transfer function.
        # For a < a_single it is GR (no early MG)
        if isinstance(z, (int, float)):
            a_single = 1/(1+z)
            a = np.linspace(a_single, 1, 100)
            # a_single is for example like for the CMB surface
        else:
            if z[0] != 0.0:
                stepsize = z[1]-z[0]
                samplesize = int(z[0]/stepsize)
                z_0_to_zmin = np.linspace(0.0, z[0] - stepsize, samplesize)
                z = np.concatenate((z_0_to_zmin, z))
            a = 1./(1.+z)
        a.sort()
        # Scale-dependant MG case with an array of k
        nk = lib.get_pk_spline_nk(cosmo.cosmo)
        status = 0
        lk, status = lib.get_pk_spline_lk(cosmo.cosmo, nk, status)
        check(status, cosmo=cosmo)
        k = np.exp(lk)
        # computing MG factor array
        mgfac_1d = 1
        mgfac_1d += _Sig_MG(cosmo, a, k)
        # converting 1D MG factor to a 2D array, so it is compatible
        # with the transfer_ka input structure in MG_add.tracer and
        # add.tracer
        mgfac_2d = mgfac_1d.reshape(len(a), -1, order='F')
        # setting transfer_ka for this case
        mg_transfer = (a, lk, mgfac_2d)

        return mg_transfer

    @unlock_instance
    def add_tracer(self, cosmo, *, kernel=None,
                   transfer_ka=None, transfer_k=None, transfer_a=None,
                   der_bessel=0, der_angles=0,
                   is_logt=False, extrap_order_lok=0, extrap_order_hik=2):
        """Adds one more tracer to the list contained in this ``Tracer``.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): cosmology object.
            kernel (:obj:`tuple`): A tuple of arrays
                ``(chi, w_chi)`` describing the radial kernel of this
                tracer. ``chi`` should contain values of the comoving
                radial distance in increasing order, and ``w_chi`` should
                contain the values of the kernel at those values of the
                radial distance. The kernel will be assumed to be zero
                outside the range of distances covered by ``chi``. If
                ``kernel`` is ``None`` a constant kernel
                :math:`W(\\chi)=1` will be assumed everywhere.
            transfer_ka (:obj:`tuple`): a tuple of arrays
                ``(a, lk,t_ka)`` describing the most general transfer
                function for a tracer. ``a`` should be an array of scale
                factor values in increasing order. ``lk`` should be an
                array of values of the natural logarithm of the wave
                number (in units of inverse Mpc) in increasing order.
                ``t_ka`` should be an array of shape ``(na, nk)``, where
                ``na`` and ``nk`` are the sizes of ``a`` and ``lk``
                respectively.``t_ka`` should hold the values of the transfer
                function at the corresponding values of ``a`` and ``lk``. If
                your transfer function is factorizable (i.e.
                :math:`T(a,k) = A(a)\\, K(k))`,
                it is more efficient to set this to ``None`` and use
                ``transfer_k`` and ``transfer_a`` to describe :math:`K` and
                :math:`A` respectively. The transfer function will be assumed
                continuous and constant outside the range of scale factors
                covered by ``a``. It will be extrapolated using polynomials
                of order ``extrap_order_lok`` and ``extrap_order_hik`` below
                and above the range of wavenumbers covered by ``lk``
                respectively. If this argument is not ``None``, the values of
                ``transfer_k`` and ``transfer_a`` will be ignored.
            transfer_k (:obj:`tuple`): a tuple of arrays
                ``(lk,t_k)`` describing the scale-dependent part of a
                factorizable transfer function. ``lk`` should be an
                array of values of the natural logarithm of the wave
                number (in units of inverse Mpc) in increasing order.
                ``t_k`` should be an array of the same size holding the
                values of the k-dependent part of the transfer function
                at those wavenumbers. It will be extrapolated using
                polynomials of order ``extrap_order_lok`` and
                ``extrap_order_hik`` below and above the range of wavenumbers
                covered by ``lk`` respectively. If ``None``, the k-dependent
                part of the transfer function will be set to 1 everywhere.
            transfer_a (:obj:`tuple`): a tuple of arrays
                `(a, t_a)`` describing the time-dependent part of a
                factorizable transfer function. ``a`` should be an array of
                scale factor values in increasing order. ``t_a`` should
                contain the time-dependent part of the transfer function
                at those values of the scale factor. The time dependence
                will be assumed continuous and constant outside the range
                covered by ``a``. If ``None``, the time-dependent part of the
                transfer function will be set to 1 everywhere.
            der_bessel (:obj:`int`): order of the derivative of the Bessel
                functions with which this tracer enters the calculation
                of the power spectrum. Allowed values are -1, 0, 1 and 2.
                0, 1 and 2 correspond to the raw functions, their first
                derivatives or their second derivatives. -1 corresponds to
                the raw functions divided by the square of their argument.
                We enable this special value because this type of dependence
                is ubiquitous for many common tracers (lensing, IAs), and
                makes the corresponding transfer functions more stable
                for small :math:`k` or :math:`\\chi`.
            der_angles (:obj:`int`): integer describing the ell-dependent prefactor
                associated with this tracer. Allowed values are 0, 1 and 2.
                0 means no prefactor. 1 means a prefactor
                :math:`\\ell(\\ell+1)`, associated with the angular
                Laplacian and used e.g. for lensing convergence and
                magnification. 2 means a prefactor
                :math:`\\sqrt{(\\ell+2)!/(\\ell-2)!}`, associated with the
                angular derivatives of spin-2 fields (e.g. cosmic shear, IAs).
            is_logt (:obj:`bool`): if ``True``, ``transfer_ka``, ``transfer_k`` and
                ``transfer_a`` will contain the natural logarithm of the
                transfer function (or their factorizable parts).
            extrap_order_lok (:obj:`int`): polynomial order used to extrapolate the
                transfer functions for low wavenumbers not covered by the
                input arrays.
            extrap_order_hik (:obj:`int`): polynomial order used to extrapolate the
                transfer functions for high wavenumbers not covered by the
                input arrays.
        """ # noqa
        is_factorizable = transfer_ka is None
        is_k_constant = (transfer_ka is None) and (transfer_k is None)
        is_a_constant = (transfer_ka is None) and (transfer_a is None)
        is_kernel_constant = kernel is None

        chi_s, wchi_s = _check_array_params(kernel, 'kernel')
        if is_factorizable:
            a_s, ta_s = _check_array_params(transfer_a, 'transfer_a')
            lk_s, tk_s = _check_array_params(transfer_k, 'transfer_k')
            tka_s = NoneArr
            if (not is_a_constant) and (a_s.shape != ta_s.shape):
                raise ValueError("Time-dependent transfer arrays "
                                 "should have the same shape")
            if (not is_k_constant) and (lk_s.shape != tk_s.shape):
                raise ValueError("Scale-dependent transfer arrays "
                                 "should have the same shape")
        else:
            a_s, lk_s, tka_s = _check_array_params(transfer_ka, 'transer_ka',
                                                   arr3=True)
            if tka_s.shape != (len(a_s), len(lk_s)):
                raise ValueError("2D transfer array has inconsistent "
                                 "shape. Should be (na,nk)")
            tka_s = tka_s.flatten()
            ta_s = NoneArr
            tk_s = NoneArr

        if not (np.diff(a_s) > 0).all():
            raise ValueError("Scale factor must be monotonically "
                             "increasing")

        status = 0
        ret = lib.cl_tracer_t_new_wrapper(cosmo.cosmo,
                                          int(der_bessel),
                                          int(der_angles),
                                          chi_s, wchi_s,
                                          a_s, lk_s,
                                          tka_s, tk_s, ta_s,
                                          int(is_logt),
                                          int(is_factorizable),
                                          int(is_k_constant),
                                          int(is_a_constant),
                                          int(is_kernel_constant),
                                          int(extrap_order_lok),
                                          int(extrap_order_hik),
                                          status)
        self._trc.append(_check_returned_tracer(ret))

    @classmethod
    def from_z_power(cls, cosmo, *, A, alpha, z_min=0., z_max=6., n_chi=1024):
        """Constructor for tracers associated with a radial kernel of the form

        .. math::
            W(\\chi) = \\frac{A}{(1+z)^\\alpha},

        where :math:`A` is an amplitude and :math:`\\alpha` is a power
        law index. The kernel only has support in the redshift range
        ``[z_min, z_max]``.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
            A (:obj:`float`): amplitude parameter.
            alpha (:obj:`float`): power law index.
            z_min (:obj:`float`): minimum redshift from to which we define the kernel.
            z_max (:obj:`float`): maximum redshift up to which we define the kernel.
            n_chi (:obj:`float`): number of intervals in the radial comoving
                distance on which we sample the kernel.
        """ # noqa
        if z_min >= z_max:
            raise ValueError("z_min should be smaller than z_max.")

        tracer = cls()

        chi_min = cosmo.comoving_radial_distance(1./(1+z_min))
        chi_max = cosmo.comoving_radial_distance(1./(1+z_max))
        chi_arr = np.linspace(chi_min, chi_max, n_chi)
        a_arr = cosmo.scale_factor_of_chi(chi_arr)
        w_arr = A * a_arr**alpha

        tracer.add_tracer(cosmo, kernel=(chi_arr, w_arr))
        return tracer

    def __del__(self):
        # Sometimes lib is freed before some Tracers, in which case, this
        # doesn't work.
        # So just check that lib.cl_tracer_t_free is still a real function.
        if hasattr(self, '_trc') and lib.cl_tracer_t_free is not None:
            for t in self._trc:
                lib.cl_tracer_t_free(t)


class NzTracer(Tracer):
    """Specific base class for tracers with an internal ``_dndz``
    redshift distribution interpolator. These include
    :func:`NumberCountsTracer` and :func:`WeakLensingTracer`.
    """

    def get_dndz(self, z):
        """Get the redshift distribution for this tracer.

        Args:
            z (:obj:`float` or `array`): redshift values.

        Returns:
            `array`: redshift distribution evaluated at the
            input values of ``z``.
        """
        return self._dndz(z)


def NumberCountsTracer(cosmo, *, dndz, bias=None, mag_bias=None,
                       has_rsd, n_samples=256):
    """Specific `Tracer` associated to galaxy clustering with linear
    scale-independent bias, including redshift-space distortions and
    magnification. The associated contributions are described in
    detail in Section 2.4.1 of the `CCL paper
    <https://arxiv.org/abs/1812.05995>`_.

    .. warning:: When including redshift-space distortions, the
        current implementation assumes linear, scale-independent growth
        Although this should be valid in :math:`\\Lambda` CDM and on the
        large scales (especially when considering a broad :math:`N(z)`),
        this approximation should be borne in mind.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        dndz (:obj:`tuple`): A tuple of arrays ``(z, N(z))``
            giving the redshift distribution of the objects. The units are
            arbitrary; ``N(z)`` will be normalized to unity.
        bias (:obj:`tuple`): A tuple of arrays ``(z, b(z))``
            giving the galaxy bias. If ``None``, this tracer won't include
            a term proportional to the matter density contrast.
        mag_bias (:obj:`tuple`): A tuple of arrays ``(z, s(z))``
            giving the magnification bias as a function of redshift. If
            ``None``, the tracer is assumed to not have magnification bias
            terms.
        has_rsd (:obj:`bool`): If ``True``, this tracer will include a
            redshift-space distortion term.
        n_samples (:obj:`int`): number of samples over which the
            magnification lensing kernel is desired. These will be equi-spaced
            in radial distance. The kernel is quite smooth, so usually O(100)
            samples is enough.
    """
    tracer = NzTracer()

    # we need the distance functions at the C layer
    cosmo.compute_distances()

    from scipy.interpolate import interp1d
    z_n, n = _check_array_params(dndz, 'dndz')
    with UnlockInstance(tracer, mutate=False):
        tracer._dndz = interp1d(z_n, n, bounds_error=False, fill_value=0)

    kernel_d = None
    if bias is not None:  # Has density term
        # Kernel
        if kernel_d is None:
            kernel_d = get_density_kernel(cosmo, dndz=dndz)
        # Transfer
        z_b, b = _check_array_params(bias, 'bias')
        # Reverse order for increasing a
        t_a = (1./(1+z_b[::-1]), b[::-1])
        tracer.add_tracer(cosmo, kernel=kernel_d, transfer_a=t_a)

    if has_rsd:  # Has RSDs
        # Kernel
        if kernel_d is None:
            kernel_d = get_density_kernel(cosmo, dndz=dndz)
        # Transfer (growth rate)
        z_b, _ = _check_array_params(dndz, 'dndz')
        a_s = 1./(1+z_b[::-1])
        t_a = (a_s, -cosmo.growth_rate(a_s))
        tracer.add_tracer(cosmo, kernel=kernel_d,
                          transfer_a=t_a, der_bessel=2)
    if mag_bias is not None:  # Has magnification bias
        # Kernel
        chi, w = get_lensing_kernel(cosmo, dndz=dndz, mag_bias=mag_bias,
                                    n_chi=n_samples)
        # Multiply by -2 for magnification
        kernel_m = (chi, -2 * w)
        if (cosmo['sigma_0'] == 0):
            # GR case
            tracer.add_tracer(cosmo, kernel=kernel_m,
                              der_bessel=-1, der_angles=1)
        else:
            # MG case
            z_b, _ = _check_array_params(dndz, 'dndz')
            tracer._MG_add_tracer(cosmo, kernel_m, z_b,
                                  der_bessel=-1, der_angles=1)
    return tracer


def WeakLensingTracer(cosmo, *, dndz, has_shear=True, ia_bias=None,
                      use_A_ia=True, n_samples=256):
    """Specific `Tracer` associated to galaxy shape distortions including
    lensing shear and intrinsic alignments within the L-NLA model.
    The associated contributions are described in detail in Section 2.4.1
    of the `CCL paper <https://arxiv.org/abs/1812.05995>`_.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        dndz (:obj:`tuple`): A tuple of arrays ``(z, N(z))``
            giving the redshift distribution of the objects. The units are
            arbitrary; ``N(z)`` will be normalized to unity.
        has_shear (:obj:`bool`): set to ``False`` if you want to omit the
            lensing shear contribution from this tracer.
        ia_bias (:obj:`tuple`): A tuple of arrays
            ``(z, A_IA(z))`` giving the intrinsic alignment amplitude
            ``A_IA(z)``. If ``None``, the tracer is assumped to not have
            intrinsic alignments.
        use_A_ia (:obj:`bool`): set to ``True`` to use the conventional IA
            normalization. Set to ``False`` to use the raw input amplitude,
            which will usually be 1 for use with perturbaion theory IA
            modeling.
        n_samples (:obj:`int`): number of samples over which the lensing
            kernel is desired. These will be equi-spaced in radial distance.
            The kernel is quite smooth, so usually O(100) samples
            is enough.
    """
    tracer = NzTracer()

    # we need the distance functions at the C layer
    cosmo.compute_distances()

    from scipy.interpolate import interp1d
    z_n, n = _check_array_params(dndz, 'dndz')
    with UnlockInstance(tracer, mutate=False):
        tracer._dndz = interp1d(z_n, n, bounds_error=False, fill_value=0)

    if has_shear:
        kernel_l = get_lensing_kernel(cosmo, dndz=dndz, n_chi=n_samples)
        if (cosmo['sigma_0'] == 0):
            # GR case
            tracer.add_tracer(cosmo, kernel=kernel_l,
                              der_bessel=-1, der_angles=2)
        else:
            # MG case
            tracer._MG_add_tracer(cosmo, kernel_l, z_n,
                                  der_bessel=-1, der_angles=2)
    if ia_bias is not None:  # Has intrinsic alignments
        z_a, tmp_a = _check_array_params(ia_bias, 'ia_bias')
        # Kernel
        kernel_i = get_density_kernel(cosmo, dndz=dndz)
        if use_A_ia:
            # Normalize so that A_IA=1
            D = cosmo.growth_factor(1./(1+z_a))
            # Transfer
            # See Joachimi et al. (2011), arXiv: 1008.3491, Eq. 6.
            # and note that we use C_1= 5e-14 from arXiv:0705.0166
            rho_m = physical_constants.RHO_CRITICAL * cosmo['Omega_m']
            a = - tmp_a * 5e-14 * rho_m / D
        else:
            # use the raw input normalization. Normally, this will be 1
            # to allow nonlinear PT IA models, where normalization is
            # already applied to the power spectrum.
            a = tmp_a
        # Reverse order for increasing a
        t_a = (1./(1+z_a[::-1]), a[::-1])
        tracer.add_tracer(cosmo, kernel=kernel_i, transfer_a=t_a,
                          der_bessel=-1, der_angles=2)
    return tracer


def CMBLensingTracer(cosmo, *, z_source, n_samples=100):
    r"""A Tracer for CMB lensing convergence :math:`\kappa`.
    The associated kernel and transfer function are described
    in Eq. 31 of the `CCL paper <https://arxiv.org/abs/1812.05995>`_.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        z_source (:obj:`float`): Redshift of source plane for CMB lensing.
        n_samples (:obj:`int`): number of samples over which the kernel
            is desired. These will be equi-spaced in radial distance.
            The kernel is quite smooth, so usually O(100) samples
            is enough.
    """
    tracer = Tracer()

    # we need the distance functions at the C layer
    cosmo.compute_distances()
    kernel = get_kappa_kernel(cosmo, z_source=z_source, n_samples=n_samples)
    if (cosmo['sigma_0'] == 0):
        tracer.add_tracer(cosmo, kernel=kernel, der_bessel=-1, der_angles=1)
    else:
        tracer._MG_add_tracer(cosmo, kernel, z_source,
                              der_bessel=-1, der_angles=1)
    return tracer


def tSZTracer(cosmo, *, z_max=6., n_chi=1024):
    """Specific :class:`Tracer` associated with the thermal Sunyaev Zel'dovich
    Compton-y parameter. The radial kernel for this tracer is simply given by

    .. math::
       W(\\chi) = \\frac{\\sigma_T}{m_ec^2} \\frac{1}{1+z},

    where :math:`\\sigma_T` is the Thomson scattering cross section and
    :math:`m_e` is the electron mass.

    Any angular power spectra computed with this tracer, should use
    a three-dimensional power spectrum involving the electron pressure
    in physical (non-comoving) units of :math:`eV\\,{\\rm cm}^{-3}`.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        z_max (:obj:`float`): maximum redshift up to which we define the
            kernel.
        n_chi (:obj:`float`): number of intervals in the radial comoving
            distance on which we sample the kernel.
    """
    # This is \sigma_T / (m_e * c^2)
    prefac = 4.01710079e-06
    return Tracer.from_z_power(cosmo, A=prefac, alpha=1, z_min=0.,
                               z_max=z_max, n_chi=n_chi)


def CIBTracer(cosmo, *, z_min=0., z_max=6., n_chi=1024):
    """Specific :class:`Tracer` associated with the cosmic infrared
    background (CIB). The radial kernel for this tracer is simply

    .. math::
       W(\\chi) = \\frac{1}{1+z}.

    Any angular power spectra computed with this tracer, should use
    a three-dimensional power spectrum involving the CIB emissivity
    density in units of
    :math:`{\\rm Jy}\\,{\\rm Mpc}^{-1}\\,{\\rm srad}^{-1}` (or
    multiples thereof -- see e.g.
    :class:`~pyccl.halos.profiles.cib_shang12.HaloProfileCIBShang12`).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        z_min (:obj:`float`): minimum redshift down to which we define the
            kernel.
        z_max (:obj:`float`): maximum redshift up to which we define the
            kernel.
        n_chi (:obj:`float`): number of intervals in the radial comoving
            distance on which we sample the kernel.
    """
    return Tracer.from_z_power(cosmo, A=1.0, alpha=1, z_min=z_min,
                               z_max=z_max, n_chi=n_chi)


def ISWTracer(cosmo, *, z_max=6., n_chi=1024):
    """Specific :class:`Tracer` associated with the integrated Sachs-Wolfe
    effect (ISW). Useful when cross-correlating any low-redshift probe with
    the primary CMB anisotropies. The ISW contribution to the temperature
    fluctuations is:

    .. math::
        \\Delta T_{\\rm CMB} =
        2T_{\\rm CMB} \\int_0^{\\chi_{LSS}}d\\chi a\\,\\dot{\\phi}

    Any angular power spectra computed with this tracer, should use
    a three-dimensional power spectrum involving the matter power spectrum.

    .. warning:: The current implementation of this tracer assumes a
        standard Poisson equation relating :math:`\\phi` and
        :math:`\\delta`, and linear, scale-independent structure growth.
        Although this should be valid in :math:`\\Lambda` CDM and on the
        large scales the ISW is sensitive to, these approximations must
        be borne in mind.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmology object.
        z_max (:obj:`float`): maximum redshift up to which we define the
            kernel.
        n_chi (:obj:`float`): number of intervals in the radial comoving
            distance on which we sample the kernel.
    """
    tracer = Tracer()

    chi_max = cosmo.comoving_radial_distance(1./(1+z_max))
    chi = np.linspace(0, chi_max, n_chi)
    a_arr = cosmo.scale_factor_of_chi(chi)
    H0 = cosmo['h'] / physical_constants.CLIGHT_HMPC
    OM = cosmo['Omega_c']+cosmo['Omega_b']
    Ez = cosmo.h_over_h0(a_arr)
    fz = cosmo.growth_rate(a_arr)
    w_arr = 3*cosmo['T_CMB']*H0**3*OM*Ez*chi**2*(1-fz)

    tracer.add_tracer(cosmo, kernel=(chi, w_arr), der_bessel=-1)
    return tracer


def _check_returned_tracer(return_val):
    """Wrapper to catch exceptions when tracers are spawned from C."""
    if (isinstance(return_val, int)):
        check(return_val)
        tr = None
    else:
        tr, _ = return_val
    return tr



import numpy as np
import pyccl as ccl
import pytest


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_massfunc_m2r_smoke(m):
    r = ccl.halos.mass2radius_lagrangian(COSMO, np.array(m))
    assert np.all(np.isfinite(r))
    assert np.shape(r) == np.shape(m)


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_sigmaM_smoke(m):
    a = 0.8
    s = ccl.sigmaM(COSMO, m, a)
    assert np.all(np.isfinite(s))
    assert np.shape(s) == np.shape(m)



__all__ = ("ConcentrationDuffy08",)

from . import Concentration


class ConcentrationDuffy08(Concentration):
    """Concentration-mass relation by `Duffy et al. 2008
    <https://arxiv.org/abs/0804.2486>`_. This parametrization is only
    valid for S.O. masses with :math:`\\Delta = \\Delta_{\\rm vir}`,
    of :math:`\\Delta=200` times the matter or critical density.
    By default it will be initialized for :math:`M_{200c}`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`): a mass
            definition object, or a name string.
    """
    name = 'Duffy08'

    def __init__(self, *, mass_def="200c"):
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name not in ["vir", "200m", "200c"]

    def _setup(self):
        vals = {"vir": (7.85, -0.081, -0.71),
                "200m": (10.14, -0.081, -1.01),
                "200c": (5.71, -0.084, -0.47)}

        self.A, self.B, self.C = vals[self.mass_def.name]

    def _concentration(self, cosmo, M, a):
        M_pivot_inv = cosmo["h"] * 5E-13
        return self.A * (M * M_pivot_inv)**self.B * a**(-self.C)
